/* 
 * File: fIMP.h
 * Author: S. Chandrasekaran
 * Date: 6.03.2014
 * Purpose: Fast indefinite multi-point clustering
 * Doc: See Documents/Papers/Clustering/cluster19.tm
 *
 * $Id: fastIMP.h,v 1.1 2014/07/31 22:35:04 shiv Exp shiv $
 *
 */

/* Functor style code.  Please #include this file. */

/* Std libraries */

#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>
#include <math.h>

/* End of Std librarirs */

/* *** *** */

/* Functor arguments */

#ifndef NS
#define UNDEFNS
#define NS(name) name
#endif

#ifndef DTYPE
#define UNDEFDTYPE
#define DTYPE float
#endif

#ifndef MAX_DTYPE
#define UNDEFMAXDTYPE
#define MAX_DTYPE FLT_MAX
#endif


/* End of Functor arguments */

/* *** *** */

/* Generate code for NS(randChoose) is needed */

#ifndef RANDCHOOSE

#include "randPerm.h"
#define RANDCHOOSE NS(randChoose)
#define UNDEFRANDCHOOSE

#endif

/* *** *** */

/* The clusters are stored in a fairly complicated way (Fortran style
   lists).  The N x L array Y contains all potential sub-cluster
   centers. The main question is which column of Y goes inside which
   cluster. The data structure needs to support efficient add, delete,
   transfer and swap. To do this I have one array, |clusterHeads|, of
   type |size_t| elements.  |clusterHeads[k]| is a |size_t| that is
   also the first column of Y that is in cluster k. The other
   sub-cluster indices are stored in the array |subClusters|. This is
   also an array of |size_t| type elements.  Here is how the
   sub-clusters are chained.  So clusterHeads[k] is the index of the
   first column of Y in cluster k.  Then subClusters[clusterHeads[k]]
   is the index of the second column of Y in cluster k, and
   subClusters[subClusters[clusterHeads[k]]] is the third column of Y
   in cluster k, and so on. The end of the cluster is reached when
   subClusters[*] == L1, which is an illegal index. Note that -1 will
   wrap around and so cannot be used.

   To make delete easier, the algorithms keep track of before and
   after pointers.  Sort of zippers C-style.

   The array T (L1 x L1) has k;l as row index but k as column
   index. Since I keep clusters compacted this will lead to an O(L^3)
   algorithm in the worst case during the combinatorial phase. To
   avoid this I store column k in column clusterHeads[k] instead. So
   clusterHeads does double duty and access to T must take this into
   account.

 */

/* *** *** */

/* 2D array reference in column-major order */
#define MREF(arr, N, i, j) arr[(j)*(N)+(i)]

/* 2D symmetric array, lower part (i >= j), in row-major order */
#define SREF(arr, i, j) arr[(((i)*((i)+1))/2)+(j)]


/* *** *** */

/* Computed euclidean distance squared between 2 columns of X */

inline DTYPE NS(euclideanDistanceX) (const size_t N,
				     const DTYPE * const restrict X,
				     const size_t i, const size_t j) {

  DTYPE dist = 0;
  size_t k;

  for (k = 0; k < N; k++) {
    DTYPE dk = MREF (X, N, k, i) - MREF (X, N, k, j);
    dist += dk * dk;
  };

  return dist;

}; /* End of NS(euclideanDistanceX) */


/* *** *** */


/* Computed euclidean distance squared between 2 columns of X & Y */

inline DTYPE NS(euclideanDistanceXY) (const size_t N,
				      const DTYPE * const restrict Y,
				      const size_t i,
				      const DTYPE * const restrict X,
				      const size_t j) {

  DTYPE dist = 0;
  size_t k;

  for (k = 0; k < N; k++) {
    DTYPE dk = MREF (Y, N, k, i) - MREF (X, N, k, j);
    dist += dk * dk;
  };

  return dist;

}; /* End of NS(euclideanDistanceXY) */


/* *** *** */


inline void NS(initializeY) (const size_t N, const size_t M, const size_t L1,
			     const size_t Q, const DTYPE * const restrict X,
			     DTYPE * const restrict Y) {

  size_t *choices = malloc (Q * sizeof (size_t));

  assert (L1 > 0);
  assert (Q <= M);

  { /* Pick first column of Y */
    size_t i;
    DTYPE dist = 0;
    size_t c = M;

    /* Pick first column of Y */

    NS(randChoose) (M, Q, choices);

    for (i = 0; i < Q; i++) {
      size_t j;
      for (j = i; j < Q; j++) {
	DTYPE dij = NS(euclideanDistanceX) (N, X, choices[i], choices[j]);
	if (dij > dist) {
	  dist = dij;
	  c = choices[i];
	};
      };
    };

    assert (dist > 0);
    assert (c < M);

    /* Y_0 = X_c */
    memcpy (Y, X + N * c, N * sizeof (DTYPE));
  };

  { /* Pick remaining columns of Y */
    size_t i;
    for (i = 1; i < L1; i++) {

      NS(randChoose) (M, Q, choices);

      DTYPE dist = 0;
      size_t c = M;
      size_t j;
      for (j = 0; j < Q; j++) {
	DTYPE dj = MAX_DTYPE;	/* Find closest existing column of Y */
	size_t k;
	for (k = 0; k < i; k++) {
	  DTYPE dk = NS(euclideanDistanceXY) (N, Y, k, X, choices[j]);
	  if (dk < dj) dj = dk;
	};
	/* Find maximum of minimum distances */
	if (dj > dist) {
	  dist = dj;
	  c = choices[j];
	};
      }; /* end for (j=0, ...) */

      assert (dist > 0 && c < M);

      /* Y_i = X_c */
      memcpy (Y + N * i, X + N * c, N * sizeof (DTYPE));
    };

  }; /* End of pick columns of Y */

  free (choices);

}; /* End of NS(initializeY) */


/* *** *** */

/* Find nearest column of Y for column X_j */
inline size_t NS(assignXj) (const size_t N, const DTYPE * const restrict X,
			    const DTYPE * const restrict Y,
			    const size_t L1, const size_t K,
			    const size_t * const restrict clusterHeads,
			    const size_t * const restrict subClusters,
			    const size_t j, size_t * const kp) {
  
  size_t k;
  size_t c = L1;
  DTYPE dist = MAX_DTYPE;

  for (k = 0; k < K; k++) {
    size_t l;
    for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
      DTYPE dl = NS(euclideanDistanceXY) (N, Y, l, X, j);
      if (dl < dist) {
	c = l;
	*kp = k;
	dist = dl;
      };
    }; /* end of for (l = clusterHeads[k] .. */
  }; /* end of for (k=0, ...) */

  assert (dist < MAX_DTYPE);
  assert (c < L1);

  return c;			/* k is returned in kp */

}; /* End of NS(assignXj) */


/* *** *** */

/* Move Y to nearest local minima by fast solver */
inline void NS(moveY) (const size_t N, const size_t M, const size_t L1,
		       const size_t L, const size_t K,
		       DTYPE * const restrict Y,
		       const DTYPE * const restrict X,
		       const DTYPE * const restrict Omega,
		       const size_t * const restrict clusterHeads,
		       const size_t * const restrict subClusters,
		       const size_t * const restrict Lambda,
		       const size_t * const restrict membership,
		       const DTYPE alpha, const DTYPE beta, const DTYPE chi,
		       const size_t * const restrict count,
		       DTYPE * const restrict D,
		       DTYPE * const restrict W,
		       DTYPE * const restrict tau,
		       DTYPE * const restrict epsilon,
		       DTYPE * const restrict mu) {

  { /* Set up D */
    size_t k;
    for (k = 0; k < K; k++) {
      size_t l;
      for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	D[l] = count[l] + alpha * Lambda[k] + chi -
	  beta * (L - Lambda[k]);
	assert (D[l] > 0);
      };
    };
  };

  DTYPE apb = alpha + beta;
  { /* Set up tau */
    size_t k;
    for (k = 0; k < K; k++) {
      tau[k] = 0;
      size_t l;
      for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	tau[k] += 1 / D[l];
      };
      assert (1.0 > apb * tau[k]);
    };
  };

  DTYPE sigma = 0;
  { /* Compute sigma */
    size_t k;
    for (k = 0; k < K; k++) sigma += tau[k] / (1 - apb * tau[k]);
  };

  { /* Set up epsilon */
    size_t k;
    for (k = 0; k < K; k++) {
      size_t l;
      for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	epsilon[l] = 1 / ((1 - apb * tau[k]) * D[l]);
      };
    };
  };

  { /* Compute 1 row of Y at a time */
    size_t i;
    for (i = 0; i < N; i++) {

      { /* Initialize and set up W =~ Xee' + chi Omega e' */
	size_t j;
	for (j = 0; j < L1; j++) W[j] = chi * Omega[i];
	for (j = 0; j < M; j++) {
	  size_t l = membership[j];
	  assert (l < L1);
	  W[l] += MREF (X, N, i, j);
	};
      };

      {  /* Y[i,:] <- psi */
	size_t k;
	for (k = 0; k < K; k++) {
	  size_t l;
	  for (l = clusterHeads[k]; l < L1; l = subClusters[l])
	    MREF (Y, N, i, l) = W[l] / D[l];
	};
      };

      { /* Compute mu from psi (Y[i,:]) */
	size_t k;
	for (k = 0; k < K; k++) {
	  size_t l;
	  mu[k] = 0;
	  for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	    mu[k] += MREF (Y, N, i, l);
	  };
	};
      };

      {	/* Compute little w and store in Y (no more psi available) */
	size_t k;
	for (k = 0; k < K; k++) {
	  size_t l;
	  for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	    MREF (Y, N, i, l) = MREF (Y, N, i, l) +
	      apb * mu[k] / ((1 - apb * tau[k]) * D[l]);
	  };
	};
      };
      
      DTYPE eW = 0;
      {	/* Compute epslion'*W  = epsilon'*z*/
	size_t k;
	for (k = 0; k < K; k++) {
	  size_t l;
	  for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	    eW += epsilon[l] * W[l];
	  };
	};
      };

      {	/* Compute Y which contains w currently */
	size_t k;
	for (k = 0; k < K; k++) {
	  size_t l;
	  for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	    MREF (Y, N, i, l) = MREF (Y, N, i, l) -
	      beta * eW * epsilon[l] / (1 + beta * sigma);
	  };
	};
      };

    }; /* end of for (i = 0; i < N; ... */
  }; /* End of compute 1 row of Y at a time */

}; /* End of NS(moveY) */


/* *** *** */

/*
  Fast Indefinite Multi-Point clustering.

  N -- dimension of space
  M -- number of samples
  L1 -- max no. of points for cluster representation
  X -- N x M column-major array of samples
  Lambda -- array of size L1; holds partitions of Y on output;
            Lambda[0], ..., Lambda[K-1]
  Y -- N x L1 array; holds sub-cluster centers on output
  alpha, beta, gamma, chi -- constants for objective function
  Omega -- N x 1 array for objective function
  Q -- number of random draws for initialization
  maxIter -- maximum number of iterations; returns actual number of iterations.

  return value = K -- actual number of clusters.

 */

size_t NS(fIMP) (const size_t N, const size_t M, const size_t L1,
		 const DTYPE * const restrict X,
		 size_t * const restrict Lambda,
		 DTYPE * const restrict Y,
		 const DTYPE alpha, const DTYPE beta, const DTYPE gamma,
		 const DTYPE chi,
		 const DTYPE * const restrict Omega,
		 const size_t Q, size_t * const restrict maxIter) {
  
  { /* Check parameters */
    assert (L1 > 1);
    assert (1 <= N);
    assert (L1 < M);
    assert (0 <= alpha);
    assert (0 <= beta && beta < chi / (2 * (L1 - 1)));
    assert (0 < chi);
    assert (0 < gamma);
    assert (1 < Q && Q <= M);
  };

  /* Initialize Y randomly (sort of) from columns of X */
  NS(initializeY) (N, M, L1, Q, X, Y);
  size_t L = L1;		/* Actual number of sub-clusters */
  size_t K = L1;		/* Actual number of clusters */

  { /* Initialize Lambda */
    size_t k;
    for (k = 0; k < L1; k++) Lambda[k] = 1;
  };

  /* clusterHeads[k] = l => Y_k;l is first sub-cluster of cluster k */
  /* The next sub-cluster in cluster k is subClusters[l] */
  size_t * const clusterHeads = malloc (L1 * sizeof (size_t));
  { /* Initialize clusterHeads */
    size_t k;
    for (k = 0; k < L1; k++) clusterHeads[k] = k;
  };

  /* subClusters[l] = p => Y_*;p is the next sub-cluster in the chain */
  /* p == L1 => end of chain */
  size_t * const subClusters = malloc (L1 * sizeof (size_t));
  { /* Initialize subClusters */
    size_t k;
    for (k = 0; k < L1; k++) subClusters[k] = L1;
  };

  /* membership[j] = k, if X_j in sub-cluster of Y_k */
  size_t * const membership = malloc (M * sizeof (size_t));

  /* count[k;l] = |S_{k;l}| */
  size_t * const count = malloc (L1 * sizeof (size_t));
  { /* Initialize count */
    size_t l;
    for (l = 0; l < L1; l++) count[l] = 0;
  };

  { /* Assign columns of X */
    size_t j;
    for (j = 0; j < M; j++) {
      size_t k;
      membership[j] =
	NS(assignXj) (N, X, Y, L1, K, clusterHeads, subClusters, j, &k);
      count[membership[j]]++;
    };
  };


  /* /\* distortion[k;l] = sum_j ||Y_k;l - X_j||^2 *\/ */
  /* DTYPE * const distortion = malloc (L1 * sizeof (DTYPE)); */
  
  /* { /\* Initialize distortion; might not be needed *\/ */
  /*   size_t j; */
  /*   for (j = 0; j < L1; j++) distortion[j] = 0; */
  /*   for (j = 0; j < M; j++) { */
  /*     size_t k = membership[j]; */
  /*     distortion[k] += NS(euclideanDistanceXY) (N, Y, k, X, j); */
  /*   }; */
  /* }; */

  /* empty[k] = true implies Y_k is not used by any sub-cluster */
  bool * const empty = malloc (L1 * sizeof (bool));
  { /* Initialize empty */
    size_t k;
    for (k = 0; k < L1; k++) empty[k] = false;
  };

#ifndef NDEBUG1
  /* Check for consistency of data structure */
  bool CheckDS (void) {
    bool passed = true;
    bool * occured = malloc (L1 * sizeof (bool));
    assert (occured != NULL);
    if (K > L) passed = false;
    if (L > L1) passed = false;
    size_t Lc = 0;
    size_t k;
    for (k = 0; k < L1; k++) occured[k] = false;
    for (k = 0; k < K; k++) {
      if (Lambda[k] <= 0) passed = false;
      Lc += Lambda[k];
      size_t l;
      size_t Kl = 0;
      for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	if (occured[l]) passed = false;
	occured[l] = true;
	Kl++;
      };
      if (Kl != Lambda[k]) passed = false;
    };
    if (Lc != L) passed = false;
    free (occured);
    if (passed) printf ("passed\n"); else printf ("failed\n");
    return passed;
  };
#endif

#ifndef NDEBUG1
  assert (CheckDS());
#endif

  DTYPE * const C = malloc (((L1 * (L1+1)) / 2) * sizeof (DTYPE));
  DTYPE * const T = malloc (L1 * L1 * sizeof (DTYPE));
  DTYPE * const R = malloc (L1 * sizeof (DTYPE));

  assert (C != NULL);
  assert (T != NULL);
  assert (R != NULL);

  /* Costs O(L^2) flops */
  inline DTYPE computeObjective (void) {
    DTYPE F = 0;

    { /* Compute distortion */
      size_t j;
      for (j = 0; j < M; j++) {
	size_t k = membership[j];
	F += NS(euclideanDistanceXY) (N, Y, k, X, j);
      };
    };

    { /* Add intra-cluster distortion */
      size_t k;
      for (k = 0; k < K; k++) {
	size_t l;
	for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	  F += chi * SREF (C, l, l);
	  size_t n;
	  for (n = clusterHeads[k]; n < L1; n = subClusters[n]) {
	    if (n < l) F += alpha * SREF (C, l, n);
	  };
	};
      };
    };

    { /* Subtract inter-cluster distortions */
      size_t k;
      for (k = 1; k < K; k++) {
	size_t p;
	for (p = 0; p < k; p++) {
	  size_t i;
	  for (i = clusterHeads[k]; i < L1; i = subClusters[i]) {
	    size_t j;
	    for (j = clusterHeads[p]; j < L1; j = subClusters[j]) {
	      DTYPE cij = (i < j) ? SREF (C, j, i) : SREF (C, i, j);
	      F += beta * (1 - gamma * cij) / gamma;
	    };
	  };
	};
      };
    };

    return F;
  }; /* End of computeObjective (void) */

  /* Allocate work space for fast solver */
  DTYPE * const W = malloc (L1 * sizeof (DTYPE));
  DTYPE * const D = malloc (L1 * sizeof (DTYPE));
  DTYPE * const tau = malloc (L1 * sizeof (DTYPE));
  DTYPE * const epsilon = malloc (L1 * sizeof (DTYPE));
  DTYPE * const mu = malloc (L1 * sizeof (DTYPE));

  assert (W != NULL);
  assert (D != NULL);
  assert (tau != NULL);
  assert (epsilon != NULL);
  assert (mu != NULL);

  /* Descending main loop */
  size_t iterCount = 0;
  bool staleY = true;		/* Y not al local minima */
  bool done = false;
  while (!done && iterCount < *maxIter) {
    
    done = true; /* Exit if nothing happens below */

#ifndef NDEBUG1
    printf ("Check at beginning ...");
    assert (CheckDS ());
    printf ("passed\n");
#endif

    { /* Initialize C (symmetric lower triangular) */
      size_t l;
      for (l = 0; l < L1; l++)
	if (!empty[l]) {
	  size_t n;
	  for (n = 0; n < l; n++)
	    if (!empty[n]) {
	      /* C[l,n] = ||Y_l - Y_n||^2 */
	      SREF (C, l, n) = NS(euclideanDistanceX) (N, Y, l, n);
	    };
	  /* C[l,l] = ||Y_l - Omega||^2 */
	  SREF (C, l, l) = NS(euclideanDistanceXY) (N, Y, l, Omega, 0);
	} else {
#ifndef NDEBUG1
	  size_t n;
	  for (n = 0; n < l; n++) {
	    /* C[l,n] = ||Y_l - Y_n||^2 */
	    SREF (C, l, n) = (DTYPE) NAN;
	  };
	  /* C[l,l] = ||Y_l - Omega||^2 */
	  SREF (C, l, l) = (DTYPE) NAN;
#endif
	};
    }; /* End of initialize C */

    { /* Initialize T */
      size_t k;
      for (k = 0; k < K; k++) {
	size_t l;
	for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	  size_t p;
	  for (p = 0; p < K; p++) {
	    if (p == k) {
	      size_t i;
	      MREF (T, L1, l, clusterHeads[k]) = 0;
	      for (i = clusterHeads[p]; i < L1; i = subClusters[i]) {
		if (i < l) {
		  MREF (T, L1, l, clusterHeads[k]) += SREF (C, l, i);
		} else if (i > l) {
		  MREF (T, L1, l, clusterHeads[k]) += SREF (C, i, l);
		};
	      };
	    } else { 		/* p != k */
	      size_t i;
	      MREF (T, L1, l, clusterHeads[p]) = 0;
	      for (i = clusterHeads[p]; i < L1; i = subClusters[i]) {
		if (i < l) {
		  MREF (T, L1, l, clusterHeads[p]) += SREF (C, l, i);
		} else {
		  assert (i != l);
		  MREF (T, L1, l, clusterHeads[p]) += SREF (C, i, l);
		};
	      }
	    };
	  };
	};
      };
    }; /* End of initialize T */

    { /* Initialize R */
      size_t k;
      for (k = 0; k < K; k++) {
	size_t l;
	for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	  R[l] = 0;
	  size_t p;
	  for (p = 0; p < k; p++)
	    R[l] += MREF (T, L1, l, clusterHeads[p]);
	  for (p = k+1; p < K; p++)
	    R[l] += MREF (T, L1, l, clusterHeads[p]);
	};
      };
    }; /* End of initialize R */

    /* Combinatorics on Y */
    {
      bool keepk;
      size_t k = 0;
      while (k < K) {

#ifndef NDEBUG1
	printf ("Check at beginning of comb loop ...");
	assert (CheckDS ());
	printf ("passed\n");
#endif

	keepk = false;		/* essentially [for] loop, unless K changes */

	size_t oldl = L1, l;
	for (l = clusterHeads[k]; l < L1; l = l < L1 ? subClusters[l] : L1) {

#ifndef NDEBUG1
	  printf ("k = %d, K = %d, L = %d, L1 = %d\n",
		  (int) k, (int) K, (int) L, (int) L1);
#endif
	  assert (Lambda[k] > 0);

	  if (count[l] == 0) { /* Delete empty cluster if possible */

	    /* del = F1 - F2 > 0 for deletion to be allowed */
	    DTYPE del = 0;
	    del = alpha * MREF (T, L1, l, clusterHeads[k]);
	    del += chi * SREF (C, l, l);
	    del += beta * (L - Lambda[k]) / gamma;
	    del -= beta * R[l];

	    if (del > 0) { /* Delete Y_k;l */
#ifndef NDEBUG1
	      printf ("Deleting Y_k;l\n");
#endif
#ifndef NDEBUG1
	      { /* Y_l = NAN */
		size_t p;
		for (p = 0; p < N; p++) MREF (Y, N, p, l) = (DTYPE) NAN;
	      };
#endif
	      empty[l] = true;
	      
	      if (oldl >= L1) {   /* First sub-cluster being deleted */

		size_t newk = subClusters[l];

		/* Move a column of T */
		if (newk < L1) { /*Cluster not empty */
		  size_t p;
		  for (p = 0; p < L1; p++) {
		    MREF (T, L1, p, newk) = MREF (T, L1, p, l);
#ifndef NDEBUG1
		    MREF (T, L1, p,l) = (DTYPE) NAN;
#endif
		  };
		};

		clusterHeads[k] = newk; /* Illegal if newk == L1 */

	      } else {	   /* Not first sub-cluster */
		subClusters[oldl] = subClusters[l];
	      };

	      subClusters[l] = L1; /* Do not point out of here. */

	      {	/* Remove Y_l from *_T_* and R_* */
		size_t q;
		for (q = 0; q < K; q++) {
		  size_t p;
		  size_t qk = clusterHeads[q];
		  for (p = qk; p < L1; p = subClusters[p]) {
		    if (p < l) {
		      MREF (T, L1, p, qk) -= SREF (C, l, p);
		      if (q != k) R[p] -= SREF (C, l, p);
		    } else if (l < p) {
		      MREF (T, L1, p, qk) -= SREF (C, p, l);
		      if (q != k) R[p] -= SREF (C, p, l);
		    } else	/* l == p */
		      assert (false);
		  };
		};
	      };

#ifndef NDEBUG1
	      {	/* Set unused entries of C, T, R, to NAN */
		size_t p;
		for (p = 0; p <= l; p++) SREF (C, l, p) = (DTYPE) NAN;
		for (p = l+1; p < L1; p++) SREF (C, p, l) = (DTYPE) NAN;
		for (p = 0; p < L1; p++) MREF (T, L1, l, p) = (DTYPE) NAN;
		R[l] = (DTYPE) NAN;
	      };
#endif

	      assert (Lambda[k] > 0);
	      (Lambda[k])--;

	      if (Lambda[k] == 0) { /* Entire cluster was deleted */

		/* assert (clusterHeads[k] >= L1); earlier illegal move? */

		/* Shift up Lambda and clusterHeads */
		size_t p;
		for (p = k; p < K-1; p++) {

		  Lambda[p] = Lambda[p+1];
		  clusterHeads[p] = clusterHeads[p+1];

		}; /* end of for (p = k; p < K-1; ...) */
		Lambda[K-1] = 0;  /* Lambda[k] > 0 assertion failing */
		clusterHeads[K-1] = L1;  /* Lambda[k] > 0 assertion failing */

		assert (K > 0);	/* K changes */
		K--;

		keepk = true;	/* don't increment k */

	      }; /* End of if (Lambda[k] == 0) */

	      assert (L > 0);
	      L--;		/* Lost one Y_l */

	      staleY = true;	/* Y not at local minima */
	      done = false;	/* Need to iterate again */

	    }; /* End of if (del > 0) */

	  }; /* End of if (count[l] == 0)... */

	  /* If done == true no empty cluster was deleted. */
	  if (done) { /* Not so fast! Choose some other move */

	    /* delta for splitting off Y_k;l */
	    DTYPE delSplitOff = -1;
	    if (Lambda[k] > 1) {
	      delSplitOff = (alpha + beta) * MREF (T, L1, l, clusterHeads[k]);
	      delSplitOff -= beta * (Lambda[k] - 1) / gamma;
	    };

	    /* delta for transferring Y_k;l; includes merging */
	    DTYPE delTransf = -1;
	    size_t transfk = L1;
	    { /* search cluster for transferring Y_k;l to */
	      size_t p;
	      for (p = 0; p < K; p++) {
		if (p != k) {
		  DTYPE del;
		  del = (alpha + beta);
		  del *= MREF (T, L1, l, clusterHeads[k]) -
		    MREF (T, L1, l, clusterHeads[p]);
		  del -= beta * ((DTYPE) Lambda[k] - Lambda[p] - 1.0) / gamma;
		  
		  if (del > delTransf) {
		    delTransf = del;
		    transfk = p;
		  };
		}; /* End of if (p != k) */
	      }; /* End of for (p = 0;) */
	    }; /* End of search clusters for transferring Y_k;l to */

	    /* delta for swapping Y_k;l */
	    DTYPE delSwap = -1;
	    size_t swapk = L1;
	    size_t swapl = L1;
	    size_t swapoldl = L1;
	    { /* Search for mini-cluster to swap with */
	      size_t p;
	      for (p = 0; p < K; p++) {
		if (p != k) {
		  size_t j;
		  size_t oldj = L1;
		  for (j = clusterHeads[p]; j < L1; j = subClusters[j]) {
		    DTYPE del;
		    del = (alpha + beta);
		    del *=
		      MREF (T, L1, l, clusterHeads[k]) +
		      MREF (T, L1, j, clusterHeads[p]) -
		      MREF (T, L1, l, clusterHeads[p]) -
		      MREF (T, L1, j, clusterHeads[k]) +
		      2 * (j < l ? SREF (C, l, j) : SREF (C, j, l));
		    
		    if (del > delSwap) {
		      delSwap = del;
		      swapk = p;
		      swapl = j;
		      swapoldl = oldj;
		    };

		    oldj = j;
		  }; /* End of for (j = clusterHeads[p]...) */
		}; /* End of if (p != k) */
	      }; /* End of for (p = 0;) */
	    }; /* End of search for mini-cluster to swap with */

	    if (delSplitOff > 0 &&
		delSplitOff >= delTransf &&
		delSplitOff >= delSwap) {
#ifndef NDEBUG1
	      printf ("Splitting off.\n");
#endif

	      assert (Lambda[k] > 1);

	      if (oldl >= L1) {	/* First sub-cluster being split-off */
		
		size_t newl = subClusters[l];
		clusterHeads[k] = newl;

		{ /* Move T from l to newl column */
		  size_t p;
		  for (p = 0; p < L1; p++)
		    MREF (T, L1, p, newl) = MREF (T, L1, p, l);
		};
		
	      } else { /* Not first sub-cluster */
		subClusters[oldl] = subClusters[l];
	      }; /* End of if (oldl >= L1) */

	      assert (K < L1);
	      clusterHeads[K] = l;
	      subClusters[l] = L1;
	      (Lambda[k])--;
	      Lambda[K] = 1;
	      K++;
	      	      
	      { /* Remove Y_k;l from *_T_k except k;l_T_k*/
		size_t i;
		for (i = 0; i < l; i++)
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, l, i);

		for (i = l+1; i < L1; i++)
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, i, l);
	      };

	      {	/* Set up *_T_l */
		size_t i;
		for (i = 0; i < l; i++) MREF (T, L1, i, l) = SREF (C, l, i);

		MREF (T, L1, l, l) = 0;

		for (i = l+1; i < L1; i++) MREF (T, L1, i, l) = SREF (C, i, l);
	      };

	      {	/* Adjust R_k;i (i != l), and R_l */
		size_t i;
		for (i = clusterHeads[k]; i < L1; i = subClusters[i])
		  if (i < l) {
		    R[i] += SREF (C, l, i);
		    R[l] += SREF (C, l, i);
		  } else {
		    assert (i != l);
		    R[i] += SREF (C, i, l);
		    R[l] += SREF (C, i, l);
		  };
	      };

	      staleY = true;	/* Y not at local minima any more */
	      done = false;	/* Make sure we come around again */

	    /* End of if (delSplitOff > 0 ...) */

	    } else if (delTransf > 0 &&
		       delTransf >= delSplitOff &&
		       delTransf >= delSwap) {
#ifndef NDEBUG1
	      printf ("Transferring.\n");
#endif

#ifndef NDEBUG1
	      printf ("Check at beginning of transfer ...");
	      assert (CheckDS ());
	      printf ("passed\n");
#endif
	      assert (transfk < K && transfk != k);

	      if (oldl >= L1) { /* First sub-cluster being transferred */

		size_t newl = subClusters[l];

		/* Move column l of T to newl */
		if (newl < L1) { /*Cluster not empty */
		  size_t p;
		  for (p = 0; p < L1; p++)
		    MREF (T, L1, p, newl) = MREF (T, L1, p, l);
		};

		clusterHeads[k] = newl;
		
	      } else { /* Not first sub-cluster */
		subClusters[oldl] = subClusters[l];
	      }; /* End of if (oldl >= L1) */

	      {	/* Cons Y_l to cluster transfk */
		size_t j = clusterHeads[transfk];
		clusterHeads[transfk] = l;
		subClusters[l] = j; /* Cannot follow subClusters[l] anymore! */

		/* Move column j of T to column l */
		{
		  size_t i;
		  for (i = 0; i < L1; i++)
		    MREF (T, L1, i, l) = MREF (T, L1, i, j);
		};
	      };

	      /* Adjust T */
	      if (clusterHeads[k] < L1) {
		/* Remove Y_k;l from *_T_k except k;l_T_k*/
		/* Add Y_k;l to *_T_transfk except l_T_transfk */
		size_t i;
		for (i = 0; i < l; i++) {
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, l, i);
		  MREF (T, L1, i, clusterHeads[transfk]) += SREF (C, l, i);
		};

		for (i = l+1; i < L1; i++) {
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, i, l);
		  MREF (T, L1, i, clusterHeads[transfk]) += SREF (C, i, l);
		};

	      } else { /* Y_l was singleton cluster */
		/* Add Y_k;l to *_T_tk except l_T_tk */
		size_t i;
		for (i = 0; i < l; i++) {
#ifndef NDEBUG1
		  MREF (T, L1, i, l) = (DTYPE) NAN;
#endif
		  MREF (T, L1, i, clusterHeads[transfk]) += SREF (C, l, i);
		};

		for (i = l+1; i < L1; i++) {
#ifndef NDEBUG1
		  MREF (T, L1, i, l) = (DTYPE) NAN;
#endif
		  MREF (T, L1, i, clusterHeads[transfk]) += SREF (C, i, l);
		};
	      };


	      /* Adjust R */
	      {	/* Adjust R_k;i */
		size_t i;
		for (i = clusterHeads[k]; i < L1; i = subClusters[i])
		  if (i < l) {
		    R[i] += SREF (C, l, i);
		    R[l] += SREF (C, l, i);
		  } else {
		    R[i] += SREF (C, i, l);
		    R[l] += SREF (C, i, l);
		  };
		/* Adjust R_transfk;i */
		for (i = clusterHeads[transfk]; i < L1; i = subClusters[i])
		  if (i < l) {
		    R[i] -= SREF (C, l, i);
		    R[l] -= SREF (C, l, i);
		  } else {
		    R[i] -= SREF (C, i, l);
		    R[l] -= SREF (C, i, l);
		  };
	      };

	      /* Adjust Lambda */
	      assert (Lambda[k] > 0);
	      (Lambda[k])--;
	      (Lambda[transfk])++;

	      if (Lambda[k] == 0) { /* One cluster disappeared */

		/* Cluster k was deleted and merged with transfk */
		{ /* Shift up Lambda and clusterHeads */
		  size_t i;
		  for (i = k; i < K-1; i++) {
		    Lambda[i] = Lambda[i+1];
		    clusterHeads[i] = clusterHeads[i+1];
		  };
		};
		Lambda[K-1] = 0;
		clusterHeads[K-1] = L1;

		assert (K > 0);
		K--;
		keepk = true;
		
	      }; /* End of if (Lambda[k] == 1) */

	      /* l is invalid now */
	      l = oldl;
	      /* New l can be invalid! So check in for loop */

	      staleY = true;
	      done = false;

#ifndef NDEBUG1
	      printf ("Check at end of transfer ...");
	      assert (CheckDS ());
	      printf ("passed\n");
#endif

	    /* End of if (delTransf > 0 ...) */

	    } else if (delSwap > 0 &&
		       delSwap >= delSplitOff &&
		       delSwap >= delTransf) {
#ifndef NDEBUG1
	      printf ("Swapping.\n");
#endif
	      assert (swapk < L1 && swapk != k);
	      assert (swapl < L1 && swapl != l);

	      /* Insert swapl into cluster k */
	      if (oldl >= L1) {
		clusterHeads[k] = swapl;
	      } else {
		subClusters[oldl] = swapl;
	      };

	      /* Insert l into cluster swapk */
	      if (swapoldl >= L1) {
		clusterHeads[swapk] = l;
	      } else {
		subClusters[swapoldl] = l;
	      };

	      /* Finish linking */
	      {
		size_t tmp = subClusters[l];
		subClusters[l] = subClusters[swapl];
		subClusters[swapl] = tmp;
	      };

	      /* Move columns of T if needed */
	      if (oldl >= L1 && swapoldl >= L1) {
		/* Swap column l and swapl of T */
		size_t i;
		for (i = 0; i < L1; i++) {
		  DTYPE tmp = MREF (T, L1, i, l);
		  MREF (T, L1, i, l) = MREF (T, L1, i, swapl);
		  MREF (T, L1, i, swapl) = tmp;
		};
	      } else if (oldl >= L1) {
		/* Move column l of T to column swapl */
		size_t i;
		for (i = 0; i < L1; i++) {
		  MREF (T, L1, i, swapl) = MREF (T, L1, i, l);
#ifndef NDEBUG1
		  MREF (T, L1, i, l) = (DTYPE) NAN;
#endif
		};
	      } else if (swapoldl >= L1) {
		/* Move column swapl of T to column l */
		size_t i;
		for (i = 0; i < L1; i++) {
		  MREF (T, L1, i, l) = MREF (T, L1, i, swapl);
#ifndef NDEBUG1
		  MREF (T, L1, i, swapl) = (DTYPE) NAN;
#endif
		};
	      }; /* End of if (oldl >= L1 && swapoldl >= L1) */

	      /* Adjust T */
	      if (Lambda[k] > 1) { /* Remove Y_l from *_T_k */
		size_t i;
		for (i = 0; i < l; i++)
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, l, i);

		for (i = l+1; i < L1; i++)
		  MREF (T, L1, i, clusterHeads[k]) -= SREF (C, i, l);
	      };

	      if (Lambda[swapk] > 1) { /* Remove Y_swapl from *_T_swapk */
		size_t i;
		for (i = 0; i < swapl; i++)
		  MREF (T, L1, i, clusterHeads[swapk]) -= SREF (C, swapl, i);

		for (i = swapl+1; i < L1; i++)
		  MREF (T, L1, i, clusterHeads[swapk]) -= SREF (C, i, swapl);
	      };

	      {	/* Add Y_l to *_T_swapk */
		size_t i;
		for (i = 0; i < l; i++)
		  MREF (T, L1, i, clusterHeads[swapk]) += SREF (C, l, i);

		for (i = l+1; i < L1; i++)
		  MREF (T, L1, i, clusterHeads[swapk]) += SREF (C, i, l);

		/* Add Y_swapl to *_T_k */
		for (i = 0; i < swapl; i++)
		  MREF (T, L1, i, clusterHeads[k]) += SREF (C, swapl, i);

		for (i = swapl+1; i < L1; i++)
		  MREF (T, L1, i, clusterHeads[k]) += SREF (C, i, swapl);
	      };

	      /* Adjust R */
	      {	/* Adjust R_k;i */
		size_t i;
		for (i = clusterHeads[k]; i < L1; i = subClusters[i]) {
		  if (i < l) {
		    R[i] += SREF (C, l, i);
		    R[l] += SREF (C, l, i);
		  } else {
		    R[i] += SREF (C, i, l);
		    R[l] += SREF (C, i, l);
		  };

		  if (i < swapl) {
		    R[i] -= SREF (C, swapl, i);
		    R[swapl] -= SREF (C, swapl, i);
		  } else {
		    R[i] -= SREF (C, i, swapl);
		    R[swapl] -= SREF (C, i, swapl);
		  };
		};

		/* Adjust R_swapk;i */
		for (i = clusterHeads[swapk]; i < L1; i = subClusters[i]) {
		  if (i < l) {
		    R[i] -= SREF (C, l, i);
		    R[l] -= SREF (C, l, i);
		  } else {
		    R[i] -= SREF (C, i, l);
		    R[l] -= SREF (C, i, l);
		  };

		  if (i < swapl) {
		    R[i] += SREF (C, swapl, i);
		    R[swapl] += SREF (C, swapl, i);
		  } else {
		    R[i] += SREF (C, i, swapl);
		    R[swapl] += SREF (C, i, swapl);
		  };
		};
	      };

	      /* l is invalid now */
	      l = swapl;

	      staleY = true;
	      done = false;

	    }; /* End of else if (delSwap > 0 ...) */

	  }; /* End of if (done) */

	  oldl = l;

	}; /* End of for (l = clusterHeads[k]; ...) */

	if (!keepk) k++;

      }; /* End of while (k < K) */

    }; /* End of Combinatorics on Y */


    if (staleY) { /* Move Y to local minima */
      NS(moveY) (N, M, L1, L, K, Y, X, Omega, clusterHeads, subClusters, Lambda,
		 membership, alpha, beta, chi, count,
		 D, W, tau, epsilon, mu);
      staleY = false;
    };

    
    { /* Re-assign X */
      size_t j;
      for (j = 0; j < M; j++) {

	bool assigned = false;

	if ((L1 > L) && (count[membership[j]] > 1)) {
	  /* Try to assign to an empty slot. */
	  size_t l = 0;
	  while (l < L1 && !empty[l]) l++;
	  assert (l < L1 && empty[l] && l != membership[j]);

	  DTYPE del = 0;
	  {
	    size_t k;
	    for (k = 0; k < K; k++) {
	      size_t i;
	      for (i = clusterHeads[k]; i < L1; i = subClusters[i])
		del += NS(euclideanDistanceXY) (N, Y, i, X, j);
	    };
	    del = beta * (del - L / gamma);
	    del += NS(euclideanDistanceXY) (N, Y, membership[j], X, j);
	    del -= chi * NS(euclideanDistanceXY) (N, Omega, 0, X, j);
	  };

	  if (del > 0) { /* Assign Y_"K;0" = Y_l = X_j */
#ifndef NDEBUG1
	    printf ("Creating new Y_k;l\n");
#endif
	    { /* Y_l = X_j */
	      size_t i;
	      for (i = 0; i < N; i++)
		MREF (Y, N, i, l) = MREF (X, N, i, j);
	    };

	    clusterHeads[K] = l;
	    subClusters[l] = L1;
	    Lambda[K] = 1;
	    empty[l] = false;
	    (count[membership[j]])--;
	    membership[j] = l;
	    count[l] = 1;
	    L++;
	    K++;

	    staleY = true;
	    assigned = true;
	    done = false;
	  };

	}; /* End of if (L1 > L) */

	if (!assigned) { /* Assign to an existing sub-cluster */
	  size_t k;
	  size_t l =
	    NS(assignXj) (N, X, Y, L1, K, clusterHeads, subClusters, j, &k);

	  if (l != membership[j]) { /* New assignment */
	    assert (count[membership[j]] > 0);
	    (count[membership[j]])--;
	    assert (!empty[l]);
	    (count[l])++;
	    membership[j] = l;
	    staleY = true;
	    done = false;
	  };
	}; /* End of if (!assigned) */

	if (staleY) { /* Move Y to local minima */
	  NS(moveY) (N, M, L1, L, K, Y, X, Omega,
		     clusterHeads, subClusters, Lambda,
		     membership, alpha, beta, chi, count,
		     D, W, tau, epsilon, mu);
	  staleY = false;
	};
	
      }; /* End of for (j = 0; j < M; ... */
    }; /* End of re-assign X */

    iterCount++;
    
  }; /* End of while(!done && iterCount < maxIter) */

  free (membership);
  free (count);
  free (C);
  free (T);
  free (R);
  free (W);
  free (D);
  free (tau);
  free (epsilon);
  free (mu);

  { /* Re-order columns of Y */
    DTYPE * tmp = malloc (N * L * sizeof (DTYPE));
    size_t ind = 0;
    size_t k;
    for (k = 0; k < K; k++) {
      size_t l;
      for (l = clusterHeads[k]; l < L1; l = subClusters[l]) {
	assert (!empty[l]);
	assert (ind < L);
	size_t i;
	for (i = 0; i < N; i++)
	  MREF (tmp, N, i, ind) = MREF (Y, N, i, l);
	ind++;
      };
    };

    memcpy (Y, tmp, N * L * sizeof (DTYPE));

#ifndef NDEBUG
    size_t tL = 0;
    for (k = 0; k < K; k++) tL += Lambda[k];
    assert (tL == L);
#endif

    free (tmp);
  };

  free (clusterHeads);
  free (subClusters);
  /* free (distortion); */
  free (empty);

  *maxIter = iterCount;
  return K;

};  /* End of NS(fIMP) */


/* *** *** */

/* Some utility routines */

/* *** *** */

/* Given l returns the cluster number k to which it belongs. Assumes
   Y_k;l are ordered by clusters, as when NS(fIMP) returns. */

inline size_t NS(clusterNo) (const size_t * const restrict Lambda,
			     const size_t K, const size_t l) {
  size_t k = 0;
  size_t l1 = l;
  while (k < K && l1 >= Lambda[k]) {
    l1 -= Lambda[k];
    k++;
  };
  return k;
}; /* End of NS(clusterNo) */

/* *** *** */

/* Returns cluster number (k) and l */
inline size_t NS(classify) (const size_t N,
			    const size_t * const restrict Lambda,
			    const size_t K, const size_t L,
			    const DTYPE * const restrict Y,
			    const DTYPE * const restrict X, size_t * const l) {
  DTYPE dist = MAX_DTYPE;
  size_t j;
  *l = L;
  for (j = 0; j < L; j++) {
    DTYPE dl = NS(euclideanDistanceXY) (N, Y, j, X, 0);
    if (dl < dist) {
      dist = dl;
      *l = j;
    };
  };

  assert (*l < L && dist < MAX_DTYPE);

  return (NS(clusterNo) (Lambda, K, *l));

}; /* End of NS(classify) */

/* *** *** */

/* Standard k-means algorithm:
   N -- dimension of space
   X -- N x M arrays of features
   M -- number of features
   Y -- on output holds K centers
   K -- number of clusters desired
   Q -- oversampling for random initialization
   maxIter -- number of iterations; returns actual no. of iterations
*/
inline void NS(kmeans) (const size_t N, const DTYPE * const restrict X,
			const size_t M,
			DTYPE * const restrict Y, const size_t K,
			const size_t Q, size_t * const restrict maxIter) {

  size_t * const membership = malloc (M * sizeof (size_t));
  size_t * const count = malloc (K * sizeof (size_t));

  assert (membership != NULL);
  assert (count != NULL);

  inline size_t assignX (size_t j) {
    DTYPE dm = MAX_DTYPE;
    size_t lm = K;
    size_t l;
    for (l = 0; l < K; l++) {
      assert (count[l] > 0);
      /* dl = ||Y_l/count[l] - X_j||^2_2 */
      DTYPE dl = 0;
      DTYPE c = count[l];
      size_t i;
      for (i = 0; i < N; i++) {
	DTYPE dt = (MREF (Y, N, i, l) / c) - MREF (X, N, i, j);
	dl +=  dt * dt ;
      };
      if (dl < dm) {
	dm = dl;
	lm = l;
      };
    };
    assert (dm < MAX_DTYPE && lm < K);
    return lm;
  };

  { /* Initialization is tricky */
    NS(initializeY) (N, M, K, Q, X, Y);
    size_t j;
    /* Temporary count for assignXj */
    for (j = 0; j < K; j++) count[j] = 1; /* for normalization only */
    /* Initialize membership */
    for (j = 0; j < M; j++) membership[j] = assignX (j);
    /* Initialize count properly */
    for (j = 0; j < K; j++) count[j] = 0;
    for (j = 0; j < M; j++) count[membership[j]]++;
#ifndef NDEBUG
    for (j = 0; j < K; j++) assert (count[j] > 0);
#endif
    /* Clear Y */
    for (j = 0; j < N * K; j++) Y[j] = 0;
    /* Re-compute Y */
    for (j = 0; j < M; j++) {
      size_t l = membership[j];
      size_t i;
      for (i = 0; i < N; i++) {
	MREF (Y, N, i, l) += MREF (X, N, i, j);
      };
    };
  }; /* End of initialization */
  /* Y[:,j] / count[j] are the centers*/

  size_t iterCount = 0;
  bool done = false;
  while (!done && iterCount < *maxIter) {
    done = true;
    size_t j;
    for (j = 0; j < M; j++) {
      size_t ln = assignX (j);
      size_t lo = membership[j];
      if (ln != lo) {	/* Membership of Xj changed */
	membership[j] = ln;
	/* Adjust Y */
	size_t i;
	for (i = 0; i < N; i++) {
	  MREF (Y, N, i, lo) -= MREF (X, N, i, j);
	  MREF (Y, N, i, ln) += MREF (X, N, i, j);
	};
	/* Adjust counts */
	assert (count[lo] > 1);	/* Yup; it cannot become empty even */
	count[lo]--;
	count[ln]++;
	/* Flag it */
	done = false;
      };
    }; /* End of for (j = 0; ... */
    iterCount++;
  }; /* end of while (!done && iterCount < maxIter) */
  *maxIter = iterCount;

  { /* Normalize Y */
    size_t l;
    for (l = 0; l < K; l++) {
      // assert (count[l] > 0); round-off errors might be the problem
      if (count[l] > 0) {
	size_t i;
	for (i = 0; i < N; i++) MREF (Y, N, i, l) /= count[l];
      } else {
	size_t i;
	for (i = 0; i < N; i++) MREF (Y, N, i, l) = 0; // = 0/0
      };
    };
  };

  free (membership);
  free (count);

}; /* End of NS(kmeans) */

/* *** *** */

#ifdef UNDEFNS
#undef NS
#undef UNDEFNS
#endif

#ifdef UNDEFDTYPE
#undef DTYPE
#undef UNDEFDTYPE
#endif

#ifdef UNDEFMAXDTYPE
#undef MAX_DTYPE
#undef UNDEFMAXDTYPE
#endif

#ifdef UNDEFRANDCHOOSE
#undef RANDCHOOSE
#undef UNDEFRANDCHOOSE
#endif

#undef MREF
#undef SREF

/* *** *** */

/* End of fIMP.h */
