/*
 * File: randPerm.h
 * Author: S. Chandrasekaran
 * Date: 4.11.2014
 * Purpose: random permutations
 *
 * $Id: randPerm.h,v 1.2 2014/04/13 15:51:13 shiv Exp shiv $
 *
 */

/* This file is meant to be #include'd */

/* There should be a typedef for DATA */

/* NS -- namespace macro */

/* You need to call srand (time (0)) yourself to randomize seed. */

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#ifndef NS
#error "Define a namespace macro NS."
#endif


/* *** *** */

/* Without -O3 consistently everywhere restrict keyword causes problems */
//#define restrict

/* *** *** */


/* Returns random number x, such that 0 <= x < N */
inline size_t NS(randRange) (const size_t N) {

  size_t rand_max = RAND_MAX;
  size_t M = N / (rand_max + 1);    /* Number of intervals - 1 */
  size_t k, j;
  size_t R;

  assert (1 < rand_max);
  assert (0 < N);
  assert (M <= rand_max); /* Limitation of code */

  /* Choose an interval randomly */
  k = rand ();
  k = k % (M + 1); /* (M * k) / rand_max */
  
  /* Find max no. of values for chosen interval */
  if (k < M) {			/* Not last interval */
    R = rand_max + 1;
  } else {			/* Last interval */
    assert (M * (rand_max + 1) <= N);
    R = N - M * (rand_max + 1);
  };

  /* Choose random number */
  j = rand ();
  j = j % R; /* (R * j) / rand_max */
  return (k * (rand_max + 1) + j);
};


/* *** *** */

/* 
 * Randomly permute [arr] of [N] elements of size [size] bytes.
 */

inline void NS(randPerm) (const size_t size, void * const arr, const size_t N) {
  size_t i;
  char tmp[size];
  // void *tmp = malloc (size);
  // assert (tmp != NULL);

  for (i = 0; i < N-1; i++) {
    size_t j = NS(randRange) (N-i);	/* Answer can be j = 0 */
    if (0 < j) {		/* Swap arr[i] with arr[i+j] */
      /* tmp = arr[i+j] */
      memcpy (tmp, arr + size * (i + j), size);
      /* arr[i+j] = arr[i] */
      memcpy (arr + size * (i + j), arr + size * i, size);
      /* arr[i] = tmp */
      memcpy (arr + size * i, tmp, size);
    };
  };

  // free (tmp);
};


/* *** *** */


/* 
 * From the numbers 0 .. N-1 choose K distinct numbers randomly.
 * Works even when N is huge and K << N.
 * Can be improved: takes O(K^2 log K) time in worst case.
 */
inline void NS(randChoose) (const size_t N, const size_t K,
			    size_t * const restrict choices) {
  size_t c;

  int cmp (const void *p1, const void *p2) {
    const size_t *t1 = p1;
    const size_t *t2 = p2;
    return ((*t1 > *t2) - (*t2 > *t1));
  };

  for (c = 0; c < K; c++) {
    size_t j = NS(randRange) (N-c);
    size_t i = 0;

    /* Find the jth number from N-c left over numbers */
    /* Assume that choices is sorted */
    i = 0;
    while (i < c && choices[i] <= j) {
      j++;			/* Since choices[i] has been removed */
      i++;
    };
    choices[c] = j;
    qsort (choices, c+1, sizeof (size_t), cmp);
  };

  return;
};


/* *** *** */


/* DEPRECATED. Use NS(randPerm) instead.
 * DATA -- type name
 * arr -- array of DATA to be randomly permuted
 * N -- size of arr
 * RANDRANGE -- macro that expands to NS(randRange)
 */

/*

#define RANDPERM(RANDRANGE, DATA, arr, N) {	\
  size_t i, j;  				\
  DATA tmp;					\
  for (i = 0; i < N-1; i++) {						\
    j = RANDRANGE (N-i);	// Answer can be j = 0		\
    if (0 < j) {		// Swap arr[i] with arr[i+j]		\
      tmp = arr[i+j];							\
      arr[i+j] = arr[i];						\
      arr[i] = tmp;							\
    };									\
  };									\
  }

*/

/* *** *** */
