/* 
 * File: testFastIMP.c
 * Author: S. Chandrasekaran
 * Date: 6.18.2014
 * Purpose: Test fastIMP.h
 *
 * $Id: testFastIMP.c,v 1.9 2014/07/25 13:46:41 shiv Exp shiv $
 *
 */


/* *** *** */

//#define restrict

/* *** *** */

/* Std includes */


#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <assert.h>
#include <stdbool.h>

/* Define reals */

typedef float Real;
#define SQRT sqrtf
#define SQ(x) (x)*(x)
#define LOG10 log10f
#define LOG2 log2f
#define POW(x,y) powf((Real)(x),(Real)(y))
#define FABS fabsf

/* *** *** */

/* Generate random code */

#define NS(name) name

#include "randPerm.h"

#undef NS

/* *** *** */

/* Make utils code (randArray, constArray, etc.) */

#define NS(name) name
#define DTYPE Real
#define ABS fabsf

#include "utils.h"

#undef NS
#undef DTYPE
#undef ABS

/* *** *** */

/* Generate fIMP code */

#define NS(name) name
#define DTYPE Real
#define MAX_DTYPE FLT_MAX
#define RANDCHOOSE randChoose

#include "fastIMP.h"

#undef RANDCHOOSE
#undef NS
#undef DTYPE
#undef MAX_DTYPE

/* *** *** */

/* Column-major order 2D array access */
#define MREF(A, M, i, j) A[(M)*(j)+(i)]

/* *** *** */


/* The X columns are grouped into K groups of M1 columns each. Each
   group is supposed to be its own cluster. For each true cluster we
   count how many fall out of the majority assigned computed
   cluster. For each computed cluster we count how many fall outside
   majority assigned true cluster. */

/* Returns no. of misclassifications */
size_t computeClassificationErrors
(const size_t K1,		/* Computed no. of clusters */
 const size_t K,		/* True no. of clusters */
 const size_t * const restrict Lambda, /* Length K1 */
 const size_t N,		      /* Dimension */
 const size_t M1,		      /* Samples per cluster */
 const Real * const restrict X,      /* Samples N * (K * M1) */
 const Real * const restrict Y) {    /* Sub-cluster centers */

  size_t L = 0;
  
  { /* Compute L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
  };

  /* Column k holds the voting pattern for cluster k. It has K1 slots. */
  size_t * const count = malloc (K1 * K * sizeof (size_t));

  { /* Fill in count */
    size_t k;
    for (k = 0; k < K1 * K; k++) count[k] = 0;
    for (k = 0; k < K; k++) {
      size_t j;
      for (j = 0; j < M1; j++) {
	size_t l, ka;
	ka = classify (N, Lambda, K1, L, Y, X + (k * M1 + j) * N, &l);
	assert (ka < K1);
	count[k * K1 + ka]++;
      };
    };
  };

  /* For each column find how many are not in voted cluster */
  size_t wrongC = 0;
  {
    size_t k;
    for (k = 0; k < K; k++) {
      size_t k1 = K1;
      size_t maxc = 0;
      size_t j;
      for (j = 0; j < K1; j++)
	if (count[k * K1 + j] > maxc) {
	  k1 = j;
	  maxc = count[k * K1 + j];
	};
      assert (k1 < K1);
      for (j = 0; j < K1; j++)
	if (j != k1) {
	  wrongC += count[k * K1 + j];
	};
    };
  };

  /* For each row find how many are not in voted cluster */
  size_t wrongR = 0;
  {
    size_t k1;
    for (k1 = 0; k1 < K1; k1++) {
      size_t k = K;
      size_t maxc = 0;
      size_t i;
      for (i = 0; i < K; i++)
	if (count[i * K1 + k1] > maxc) {
	  k = i;
	  maxc = count[i * K1 + k1];
	};
      if (k < K && maxc > 0) {		/* Cluster was not empty */
	for (i = 0; i < K; i++)
	  if (i != k) {
	    wrongR += count[i * K1 + k1];
	  };
      };
    };
  };

  /* Free memory */
  free (count);

  return (wrongC + wrongR);

}; /* End of computeClassificationErrors */


/* *** *** */

/* Generalization of computeClassificationErrors */
size_t computeClassificationErrorsG
(const size_t KY,		/* Computed no. of clusters */
 const size_t KX,		/* True no. of clusters */
 const size_t * const restrict LambdaY, /* Length KY, partition of Y */
 const size_t N,		      /* Dimension */
 const size_t * const restrict LambdaX, /* Length KX, partition of X */
 const Real * const restrict X,      /* Samples N * sum LambdaX[] */
 const Real * const restrict Y,
 size_t * wrongC,		/* output column-wise error count */
 size_t * wrongR		/* output row-wise error count */
) {    /* Sub-cluster centers */

  size_t LY = 0;		/* no. of columns of Y */
  
  { /* Compute LY */
    size_t k;
    for (k = 0; k < KY; k++) LY += LambdaY[k];
  };

  size_t LX = 0;		/* no. of columns of X */
  
  { /* Compute LX */
    size_t k;
    for (k = 0; k < KX; k++) LX += LambdaX[k];
  };

  /* Column k holds the voting pattern for cluster k. It has KY slots. */
  size_t * const count = malloc (KY * KX * sizeof (size_t));

  { /* Fill in count */
    size_t k;
    for (k = 0; k < KY * KX; k++) count[k] = 0;
    size_t offset = 0;
    for (k = 0; k < KX; k++) {
      size_t j;
      for (j = 0; j < LambdaX[k]; j++) {
	size_t l, ka;
	assert (offset + j < LX);
	ka = classify (N, LambdaY, KY, LY, Y, X + (offset + j) * N, &l);
	assert (ka < KY);
	count[k * KY + ka]++;
      };
      offset += LambdaX[k];
    };
  };

  /* For each column find how many are not in voted cluster */
  *wrongC = 0;
  {
    size_t k;
    for (k = 0; k < KX; k++) {
      size_t ky = KY;
      size_t maxc = 0;
      size_t j;
      for (j = 0; j < KY; j++)
	if (count[k * KY + j] > maxc) {
	  ky = j;
	  maxc = count[k * KY + j];
	};
      assert (ky < KY);
      for (j = 0; j < KY; j++)
	if (j != ky) {
	  *wrongC += count[k * KY + j];
	};
    };
  };

  /* For each row find how many are not in voted cluster */
  *wrongR = 0;
  {
    size_t ky;
    for (ky = 0; ky < KY; ky++) {
      size_t kx = KX;
      size_t maxc = 0;
      size_t i;
      for (i = 0; i < KX; i++)
	if (count[i * KY + ky] > maxc) {
	  kx = i;
	  maxc = count[i * KY + ky];
	};
      if (kx < KX && maxc > 0) {		/* Cluster was not empty */
	for (i = 0; i < KX; i++)
	  if (i != kx) {
	    *wrongR += count[i * KY + ky];
	  };
      };
    };
  };

  /* Free memory */
  free (count);

  return (*wrongC + *wrongR);

}; /* End of computeClassificationErrorsG */


/* *** *** */

/* Automatically choose Q=K for K-means runs */
void driver (const size_t N, const size_t K, const size_t M1, const size_t Ks,
	     const Real gamma, const Real chi, const Real alpha,
	     const size_t Q, const size_t MaxIter,
	     Real * const restrict scoreIMP,
	     Real * const restrict scoreKMeans,
	     size_t * const restrict impCount,
	     size_t * const restrict kmeansCount,
	     Real * timeIMP,
	     Real * timeKMeans) {
  
  size_t M = M1 * K;		/* no. of samples */
  size_t L1 = K * Ks;		/* max number of sub-clusters */

  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);
  randArray (X, N * M);

  /* Return random number in [-b,b] */
  inline Real randomReal (Real b) {
    return b * (2 * drand48 () - 1);
  };

  //const Real nu = 1.5 * K;  /* Inflation parameter for clusters */
  const Real nu = 10.0 * LOG2(K) * POW (K, 1.0/N);
  Real * const shifts = malloc (N * K * sizeof (Real));  /* True shifts */
  assert (shifts != NULL);

  { /* Set up shifts and shift clusters */
    size_t k;

    randArray (shifts, N * K);	/* Random shifts */

    /* Random scaling & shift */
    for (k = 0; k < K; k++) {
      scaleColumn (N, shifts, k, randomReal (nu)); /* random scaling */
      shiftMatrix (shifts + k * N, X + k * N * M1, N, M1);
    };

    //printMatrix (shifts, N, K);
  };

  Real * const Xs = malloc (N * M * sizeof (Real));
  /* Xs = X */
  memcpy (Xs, X, N * M * sizeof (Real));
  /* Scramble columns of Xs */
  randPerm (N * sizeof (Real), Xs, M);


  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  Real beta = chi / (2.00001 * (L1 - 1));
  *impCount = MaxIter;	/* Return value is actual count */
  size_t K1;			/* Returned value of K */
  clock_t start, end;
  start = clock ();
  K1 = fIMP (N, M, L1, Xs, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, impCount);
  end = clock ();
  *timeIMP = ((double) (end - start)) / CLOCKS_PER_SEC;
  size_t L = 0;
  { /* Compute L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
  };

  /* Score based on Y */
  size_t wrongY;
  wrongY = computeClassificationErrors (K1, K, Lambda, N, M1, X, Y);

  size_t wrongS;
  { /* Score based on shifts */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongS = computeClassificationErrors (K, K, Lambda1, N, M1, X, shifts);
    free (Lambda1);
  };

  *scoreIMP = ((double) wrongS - wrongY) / (2 * M);

  { /* Standard K-Means using Q = K*/
    clock_t start, end;
    *kmeansCount = MaxIter;
    start = clock ();
    kmeans (N, Xs, M, Y, K, K, kmeansCount);
    end = clock ();
    *timeKMeans = ((double) (end - start)) / CLOCKS_PER_SEC;
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongK = computeClassificationErrors (K, K, Lambda1, N, M1, X, Y);
    free (Lambda1);
  };

  *scoreKMeans = ((double) wrongS - wrongK) / (2 * M);

  /* Free memory */
  free (X);
  free (shifts);
  free (Xs);
  free (Y);
  free (Lambda);
  free (Omega);

}; /* End of driver */


/* *** *** */


/* *** Tests *** */


/* *** *** */

/* Primarily visualization in 2D */
void test1 (void) {

  size_t N = 7;
  size_t M1 = 30;			/* samples per cluster */
  size_t K = 100;			/* actual number of clusters */
  size_t M = M1 * K;			/* no. of samples */
  size_t Ks = 2;			/* sub-clusters per cluster */
  size_t L1 = K * Ks;;

  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);
  randArray (X, N * M);

  /* Return random number in [-b,b] */
  inline Real randomReal (Real b) {
    return b * (2 * drand48 () - 1);
  };

  // const Real nu = 1.5 * K;	  /* Inflation parameter for clusters */
  const Real nu = 10.0 * LOG2(K) * POW(K, 1.0 / N);
  printf ("nu = %g\n", (double) nu);
  Real * const shifts = malloc (N * K * sizeof (Real));  /* True shifts */
  assert (shifts != NULL);

  { /* Set up shifts and shift clusters */
    size_t k;

    randArray (shifts, N * K);	/* Random shifts */

    /* Random scaling & shift */
    for (k = 0; k < K; k++) {
      scaleColumn (N, shifts, k, randomReal (nu)); /* random scaling */
      shiftMatrix (shifts + k * N, X + k * N * M1, N, M1);
    };

    //printMatrix (shifts, N, K);
  };

  Real * const Xs = malloc (N * M * sizeof (Real));
  /* Xs = X */
  memcpy (Xs, X, N * M * sizeof (Real));
  /* Scramble columns of Xs */
  randPerm (N * sizeof (Real), Xs, M);


  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  /* fIMP parameters */
  /* const Real gamma = 1E-3;
  const Real chi = 0.03;
  const Real beta = chi / (2.00001 * (L1 - 1));
  const Real alpha = 0.3 / (LOG2(N) * SQRT(K * Ks)); */
  const Real chi1 = (Real) M1 / ((Real) Ks); /* M/L */
  const Real chi = 0.001 * chi1;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  //const Real gamma = chi1 * (Real) Ks / (M1 * 2.0);
  const Real gamma = 4E-4;	/* O(1) */
  //const Real alpha = 0.0001 * (Real) M1 / ((Real) Ks);
  //const Real alpha = 0.001 * chi1;
  const Real alpha = 2 * (L1 - 1) * beta;	/* M/L */
  const size_t Q = 30; /* K * (2 + M1 / 5); */
  const size_t MaxIter = 300;
  size_t maxIter = MaxIter; /* Return value of iteration count is stored here */

  size_t K1;			/* Returned value of K */
  K1 = fIMP (N, M, L1, Xs, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, &maxIter);
  printf ("Iterations = %d\n", (int) maxIter);
  printf ("True K = %d computed K = %d\n", (int) K, (int) K1);

  size_t L = 0;
  { /* Compute and print L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
    printf ("True L = %d computed L = %d\n", (int) L1, (int) L);
  };

  size_t wrongY;
  { /* Score based on Y */
    wrongY = computeClassificationErrors (K1, K, Lambda, N, M1, X, Y);
    printf ("Wrong for computed Y = %d out of %d\n", (int) wrongY, (int) M);
  };

  size_t wrongS;
  { /* Score based on shifts */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongS = computeClassificationErrors (K, K, Lambda1, N, M1, X, shifts);
    printf ("Wrong for true shifts = %d out of %d\n", (int) wrongS, (int) M);
    free (Lambda1);
  };

  printf ("Final score = %g\n", ((double) wrongS - wrongY) / (2 * M));

  /* Visualize clusters */
  if (N == 2) {
    FILE *gplotout;

    gplotout = popen ("gnuplot -persist", "w");
    if (gplotout != NULL) {	/* We have gnuplot */

      /* Gnuplot color table */
      char *gpc[21];
      gpc[0] = "black";
      gpc[1] = "red";
      gpc[2] = "brown";
      gpc[3] = "green";
      gpc[4] = "blue";
      gpc[5] = "gray";
      gpc[6] = "greenyellow";
      gpc[7] = "yellow";
      gpc[8] = "plum";
      gpc[9] = "bisque";
      gpc[10] = "violet";
      gpc[11] = "orange";
      gpc[12] = "pink";
      gpc[13] = "honeydew";
      gpc[14] = "cyan";
      gpc[15] = "magenta";
      gpc[16] = "gold";
      gpc[17] = "salmon";
      gpc[18] = "khaki";
      gpc[19] = "beige";
      gpc[20] = "olive";
      
      /* *** *** */

      /* Set up gnuplot */
      bool pdf = false;
      if (pdf) {
	fprintf (gplotout, "set terminal pdf color\n");
	fprintf (gplotout, "set output \"fIMP.pdf\"\n");
      };
      fprintf (gplotout, "set multiplot\n");
      fprintf (gplotout, "set parametric\n");
      fprintf (gplotout, "set xrange [-%g:%g]\n", nu+1, nu+1);
      fprintf (gplotout, "set yrange [-%g:%g]\n", nu+1, nu+1);
      //fprintf (gplotout, "set size ratio -1\n");
      fprintf (gplotout, "unset xtics\n");
      fprintf (gplotout, "unset ytics\n");
      fprintf (gplotout, "set title \"fIMP\"\n");

      size_t k, j;

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "set object circle at %g,%g radius 0.05\n",
		 MREF (shifts, N, 0, k), MREF (shifts, N, 1, k));
      };

      { /* Draw computed centers */
	size_t ind = 0;
	for (k = 0; k < K1; k++) {
	  size_t l;
	  /* printf ("%d \"%s\" sub-cluster(s)\n", (int) Lambda[k], gpc[k]); */
	  for (l = 0; l < Lambda[k]; l++) {
	    fprintf (gplotout,
		     "set object rectangle at %g,%g size 0.1, 0.1 fc rgb \"%s\"\n",
		     MREF (Y, N, 0, ind), MREF (Y, N, 1, ind), gpc[k]);
	    ind++;
	  };
	};
      };

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "plot '-' w p lt %d\n", (int) (k+1));
	for (j = 0; j < M1; j++) {
	  fprintf (gplotout, "%g %g\n",
		   (double) MREF (X, N, 0, j+k*M1),
		   (double) MREF (X, N, 1, j+k*M1));
	};
	fprintf (gplotout, "e\n");
      };

      if (pdf) fprintf (gplotout, "set output\n");
      fflush (gplotout);
      
    };
  }; /* End of visualize clusters */

  { /* Standard K-Means with Q = K */
    maxIter = MaxIter;		/* reset */
    kmeans (N, Xs, M, Y, K, K, &maxIter);
    printf ("kmeans took %d iterations\n", (int) maxIter);
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongK = computeClassificationErrors (K, K, Lambda1, N, M1, X, Y);
    printf ("Wrong for K-means = %d out of %d\n", (int) wrongK, (int) M);
    free (Lambda1);
  };

  printf ("Final score for K-means = %g\n",
	  ((double) wrongS - wrongK) / (2 * M));

  /* Visualize kmeans clusters */
  if (N == 2) {
    FILE *gplotout;

    gplotout = popen ("gnuplot -persist", "w");
    if (gplotout != NULL) {	/* We have gnuplot */

      /* Gnuplot color table */
      char *gpc[21];
      gpc[0] = "black";
      gpc[1] = "red";
      gpc[2] = "brown";
      gpc[3] = "green";
      gpc[4] = "blue";
      gpc[5] = "gray";
      gpc[6] = "greenyellow";
      gpc[7] = "yellow";
      gpc[8] = "plum";
      gpc[9] = "bisque";
      gpc[10] = "violet";
      gpc[11] = "orange";
      gpc[12] = "pink";
      gpc[13] = "honeydew";
      gpc[14] = "cyan";
      gpc[15] = "magenta";
      gpc[16] = "gold";
      gpc[17] = "salmon";
      gpc[18] = "khaki";
      gpc[19] = "beige";
      gpc[20] = "olive";
      
      /* *** *** */

      /* Set up gnuplot */
      bool pdf = false;
      if (pdf) {
	fprintf (gplotout, "set terminal pdf color\n");
	fprintf (gplotout, "set output \"kmeans.pdf\"\n");
      };
      fprintf (gplotout, "set multiplot\n");
      fprintf (gplotout, "set parametric\n");
      fprintf (gplotout, "set xrange [-%g:%g]\n", nu+1, nu+1);
      fprintf (gplotout, "set yrange [-%g:%g]\n", nu+1, nu+1);
      //fprintf (gplotout, "set size ratio -1\n");
      fprintf (gplotout, "unset xtics\n");
      fprintf (gplotout, "unset ytics\n");
      fprintf (gplotout, "set title \"kmeans\"\n");

      size_t k, j;

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "set object circle at %g,%g radius 0.05\n",
		 MREF (shifts, N, 0, k), MREF (shifts, N, 1, k));
      };

      /* Draw computed centers */
      for (k = 0; k < K; k++) {
	fprintf (gplotout,
		 "set object rectangle at %g,%g size 0.1, 0.1 fc rgb \"%s\"\n",
		 MREF (Y, N, 0, k), MREF (Y, N, 1, k), gpc[k]);
      };

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "plot '-' w p lt %d\n", (int) (k+1));
	for (j = 0; j < M1; j++) {
	  fprintf (gplotout, "%g %g\n",
		   (double) MREF (X, N, 0, j+k*M1),
		   (double) MREF (X, N, 1, j+k*M1));
	};
	fprintf (gplotout, "e\n");
      };

      if (pdf) fprintf (gplotout, "set output\n");
      fflush (gplotout);
      
    };
  }; /* End of visualize kmeans clusters */

  /*  Memory clean up */
  free (X);
  free (Xs);
  free (shifts);
  free (Y);
  free (Lambda);
  free (Omega);

}; /* End of test1 */

/* *** *** */

/* Gather error statistics. Repeatedly run tests and gather statistics
   on scores and iterations counts. */
void test2 () {

  /* nu = 1.5 * K

     Good values for fIMP when K = 100, Ks = 2, chi = 0.01:
     N =   7: gamma = 7E-4; alpha = 5E-3
     N =  14: gamma = 7E-4; alpha = 5E-3
     N =  28: gamma = 4E-4; alpha = 5E-3
     N =  56: gamma = 2E-4; alpha = 4E-3 
     N = 112: gamma = 1E-4; alpha = 4E-3

     Good values for fIMP when K = 200, Ks = 2, chi = 0.03:
     N =   7: gamma = 7E-4; alpha = 5E-3
     N =  14: gamma = 7E-4; alpha = 5E-3
     N =  28: gamma = 4E-4; alpha = 5E-3
     N =  56: gamma = 2E-4; alpha = 4E-3 
     N = 112: gamma = 1E-4; alpha = 4E-3
  */

  /* nu = 10.0 * LOG2(K) * POW(K,1/N)

     K = 100, Ks = 2 Q = 30
     N = 7, gamma = 4E-4 chi = 0.001 * chi1 M1 = 30
     N = 14, gamma = 4E-4 chi = 0.001 * chi1 M1 = 30
     N = 28, gamma = 4E-4, chi = 0.001 * chi1 M1 = 60
     N = 56, gamma = 3E-4, chi = 0.001 * chi1 M1 = 120
     N = 112, gamma = 3E-4, chi = 0.001 * chi1 M1 = 240

     K = 200, Ks = 2 Q = 40
     N = 7, gamma = 2.5E-4 chi = 0.001 * chi1 M1 = 30
     N = 14, gamma = 2E-4 chi = 0.001 * chi1 M1 = 30
     N = 28, gamma = 2E-4, chi = 0.001 * chi1 M1 = 60
     N = 56, gamma = 2E-4, chi = 0.001 * chi1 M1 = 120
     N = 112, gamma = 8E-5, chi = 0.001 * chi1 M1 = 240
 */

  /* Fixed parameters of test */
  const size_t Trials = 50;		/* Number of trials */
  const size_t N = 112;			/* dimension */
  const size_t K = 200;			/* number of clusters */
  const size_t M1 = 240;			/* Samples per cluster */
  const size_t Ks = 2;			/* Sub-clusters per cluster */
  const Real chi1 = (Real) M1 / Ks;	/* M/L */
  const Real chi = 0.001 * chi1;
  const Real gamma = 8E-5;	/* O(1) */
  // const Real gamma = 3E-4;
  const size_t L1 = K * Ks;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  const Real alpha = 2 * (L1 - 1) * beta;		/* M/L */
  // const Real alpha = 5E-3;
  const size_t Q = 40; /* M1 for fIMP */ /* K * (2 + M1 / 5); */
  const size_t MaxIter = 300;

  /* Output variables for driver */
  Real scoreIMP = 0;
  Real scoreKMeans = 0;
  size_t impCount = MaxIter;
  size_t kmeansCount = MaxIter;
  Real timeIMP;
  Real timeKMeans;

  size_t trial;
  size_t wins = 0, absWins = 0;
  size_t losses = 0, absLosses = 0;
  size_t ties = 0, absTies = 0;
  Real avgScoreIMP = 0;
  Real avgScoreKMeans = 0;
  Real avgTimeIMP = 0;
  Real avgTimeKMeans = 0;
  for (trial = 0; trial < Trials; trial++) {
    printf ("\b\b\b\b\b\b\b\b\b\b\b\b\b\btrial = %d", (int) trial);
    fflush (stdout);
    driver (N, K, M1, Ks, gamma, chi, alpha, Q, MaxIter,
	    &scoreIMP, &scoreKMeans, &impCount, &kmeansCount,
	    &timeIMP, &timeKMeans);
    if (impCount >= MaxIter) printf ("IMP requires more iterations\n");
    if (kmeansCount >= MaxIter) printf ("KMeans requires more iterations\n");
    if (scoreIMP > scoreKMeans)
      wins++;
    else if (scoreIMP < scoreKMeans)
      losses++;
    else
      ties++;
    if (scoreIMP > 0)
      absWins++;
    else if (scoreIMP < 0)
      absLosses++;
    else
      absTies++;
    avgScoreIMP += scoreIMP;
    avgScoreKMeans += scoreKMeans;
    avgTimeIMP += timeIMP;
    avgTimeKMeans += timeKMeans;
  };

  printf ("\nN = %d Q = %d M1 = %d K = %d Ks = %d Trials = %d\n",
	  (int) N, (int) Q, (int) M1, (int) K, (int) Ks, (int) Trials);
  printf ("chi = %g alpha = %g beta = %g gamma = %g\n",
	  (double) chi, (double) alpha, (double) beta, (double) gamma);
  printf ("AbsWins = %d AbsTies = %d AbsLosses = %d\n",
	  (int) absWins, (int) absTies, (int) absLosses);
  printf ("Wins = %d Ties = %d Losses = %d\n",
	  (int) wins, (int) ties, (int) losses);
  printf ("Avg. score of IMP = %g\n", (double) avgScoreIMP / Trials);
  printf ("Avg. score of KMeans = %g\n", (double) avgScoreKMeans / Trials);
  printf ("Avg. time of IMP = %g\n", (double) avgTimeIMP / Trials);
  printf ("Avg. time of KMeans = %g\n", (double) avgTimeKMeans / Trials);

}; /* End of test2 */

/* *** *** */

/* Run tuned tests for synthetic data for paper */
void test3 (void) {
  
  const size_t Ks = 2;
  const size_t Ns[] = {7, 14, 28, 56, 112};
  const size_t NN = 5;
  const size_t MaxIter = 300;
  const size_t M1 = 30;

  void metaDriver (const size_t K, const Real chi,
		   const Real * const restrict gammas,
		   const Real * const restrict alphas,
		   const size_t Q,
		   const size_t Trials) {
    size_t nn;
    for (nn = 0; nn < NN; nn++) {
      printf ("\nN = %d K = %d chi = %g gamma = %g alpha = %g\n",
	      (int) Ns[nn], (int) K, (Real) chi, 
	      (double) gammas[nn], (double) alphas[nn]);
      /* Output variables for driver */
      Real scoreIMP = 0;
      Real scoreKMeans = 0;
      size_t impCount = MaxIter;
      size_t kmeansCount = MaxIter;
      Real timeIMP;
      Real timeKMeans;
      
      size_t trial;
      size_t wins = 0, absWins = 0;
      size_t losses = 0, absLosses = 0;
      size_t ties = 0, absTies = 0;
      Real avgScoreIMP = 0;
      Real avgScoreKMeans = 0;
      Real avgTimeIMP = 0;
      Real avgTimeKMeans = 0;
      size_t itsIMP = 0;
      size_t itsKMeans = 0;
      for (trial = 0; trial < Trials; trial++) {
	// printf ("\b\b\b\b\b\b\b\b\b\b\b\b\b\btrial = %d", (int) trial);
	driver (Ns[nn], K, M1, Ks, gammas[nn], chi, alphas[nn], Q, MaxIter,
		&scoreIMP, &scoreKMeans, &impCount, &kmeansCount,
		&timeIMP, &timeKMeans);
	if (impCount >= MaxIter) printf ("IMP requires more iterations\n");
	if (kmeansCount >= MaxIter) printf ("KMeans requires more iterations\n");

	if (scoreIMP > scoreKMeans) wins++;
	else if (scoreIMP < scoreKMeans) losses++;
	else ties++;

	if (scoreIMP > 0) absWins++;
	else if (scoreIMP < 0) absLosses++;
	else absTies++;

	avgScoreIMP += scoreIMP;
	avgScoreKMeans += scoreKMeans;
	avgTimeIMP += timeIMP;
	avgTimeKMeans += timeKMeans;
	itsIMP += impCount;
	itsKMeans += kmeansCount;
      };
      
      printf ("AbsWins = %d AbsTies = %d AbsLosses = %d\n",
	      (int) absWins, (int) absTies, (int) absLosses);
      printf ("Wins = %d Ties = %d Losses = %d\n",
	      (int) wins, (int) ties, (int) losses);
      printf ("average score IMP = %g\n", (double) avgScoreIMP / Trials);
      printf ("average score KMeans = %g\n", (double) avgScoreKMeans / Trials);
      printf ("average iterations IMP = %g\n", (double) itsIMP / Trials);
      printf ("average iterations KMeans = %g\n", (double) itsKMeans / Trials);
      printf ("average time for IMP = %g\n", (double) avgTimeIMP / Trials);
      printf ("average time for KMeans = %g\n", (double) avgTimeKMeans / Trials);
      fflush (stdout);
    }; /* End of for (nn = 0; nn < NN... */
  }; /* End of MetaDriver */
  
  {
    const size_t K = 100;
    const Real chi1 = (Real) M1 / Ks;
    const Real chi = 0.001 * chi1;
    const Real const gammas[] =
      {(Real) 4E-4, (Real) 3E-4, (Real) 2E-4, (Real) 1E-4, (Real) 5E-5};
    const size_t L1 = K * Ks;
    const Real beta = chi / (2.00001 * (L1 - 1));
    const Real Alpha = 2 * (L1 - 1) * beta;
    const Real const alphas[] = {Alpha, Alpha, Alpha, Alpha, Alpha};
    const size_t Q = 30;
    const size_t Trials = 100;

    metaDriver (K, chi, gammas, alphas, Q, Trials);
  };

  {
    const size_t K = 200;
    const Real chi1 = (Real) M1 / Ks;
    const Real chi = 0.001 * chi1;
    const Real const gammas[] =
      {(Real) 9E-5, (Real) 9E-5, (Real) 7E-5, (Real) 5E-5, (Real) 4E-5};
    const size_t L1 = K * Ks;
    const Real beta = chi / (2.00001 * (L1 - 1));
    const Real Alpha = 2 * (L1 - 1) * beta;
    const Real const alphas[] = {Alpha, Alpha, Alpha, Alpha, Alpha};
    const size_t Q = 45;
    const size_t Trials = 50;

    metaDriver (K, chi, gammas, alphas, Q, Trials);
  };

  fprintf (stderr, "Done\n");
}; /* End of test3 */

/* *** *** */

/* 
   KMeans consistently returns 12 "wrong"
   fIMP's best answer was 10 "wrong", though 12 is more consistent.
   This is with row scaling thrown in. Without that the errors are much worse all
   around.
 */
void test4 (void) {
  
  const size_t M = 150;		/* No. of samples (hand counted) */
  const size_t N  = 4;		/* as documented */
  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);

  const size_t K = 3;
  /* size_t irisVirginica = 50;
     size_t irisVersicolor = 50;
     size_t irisSetosa = 50; */
  const size_t M1 = 50;		/* no. of samples per class */
  const size_t Ks = 2;
  const size_t L1 = K * Ks;

  { /* Read X */
    FILE * file = fopen ("iris.data", "r");
    size_t i;
    char s[30];
    for (i = 0; i < M; i++) {
      fscanf (file, "%f%*c%f%*c%f%*c%f%*c%s\n",
	      X + i * N + 0, X + i * N + 1, X + i * N + 2, X + i * N + 3, s);
    };
    int err = fclose (file);
    assert (err == 0);
  };

  { /* Scale each row to [0,1] */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Real maxx = 0;
      for (j = 0; j < M; j++) {
	size_t Xij = FABS(X[j * N + i]);
	if (Xij > maxx) maxx = Xij;
      };
      if (maxx > 0) {
	for (j = 0; j < M; j++) {
	  X[j*N+i] /= maxx;
	};
      };
    }
  };

  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  /* fIMP parameters */
  const Real chi1 = (Real) M1 / ((Real) Ks); /* M/L */
  const Real chi = 0.0001 * chi1;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  const Real gamma = 0.7;	/* O(1) */
  const Real alpha = 2.0001 * (L1 - 1) * beta;	/* M/L */
  const size_t Q = K * (2 + M1 / 5);
  const size_t MaxIter = 300;
  size_t maxIter = MaxIter; /* Return value of iteration count is stored here */

  size_t K1;			/* Returned value of K */
  K1 = fIMP (N, M, L1, X, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, &maxIter);
  printf ("Iterations = %d\n", (int) maxIter);
  printf ("True K = %d computed K = %d\n", (int) K, (int) K1);

  size_t L = 0;
  { /* Compute and print L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
    printf ("True L = %d computed L = %d\n", (int) L1, (int) L);
  };

  size_t wrongY;
  { /* Score based on Y */
    wrongY = computeClassificationErrors (K1, K, Lambda, N, M1, X, Y);
    printf ("Wrong for computed Y = %d out of %d\n", (int) wrongY, (int) M);
  };

  { /* Standard K-Means */
    maxIter = MaxIter;		/* reset */
    kmeans (N, X, M, Y, K, Q, &maxIter);
    printf ("kmeans took %d iterations\n", (int) maxIter);
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongK = computeClassificationErrors (K, K, Lambda1, N, M1, X, Y);
    printf ("Wrong for K-means = %d out of %d\n", (int) wrongK, (int) M);
    free (Lambda1);
  };

  free (X);
  free (Y);
  free (Lambda);
  free (Omega);

}; /* End of test4 */

/* *** *** */

void test5 (void) {
  
  /* Color RGB codes */
  const Real BrownR = 165;
  const Real BrownG = 42;
  const Real BrownB = 42;
  const Real BuffR = 240;
  const Real BuffG = 220;
  const Real BuffB = 130;
  const Real CinnamonR = 210;
  const Real CinnamonG = 105;
  const Real CinnamonB = 30;
  const Real ChocolateR = 78;
  const Real ChocolateG = 46;
  const Real ChocolateB = 40;
  const Real GrayR = 192;
  const Real GrayG = 192;
  const Real GrayB = 192;
  const Real GreenR = 0;
  const Real GreenG = 255;
  const Real GreenB = 0;
  const Real PinkR = 255;
  const Real PinkG = 192;
  const Real PinkB = 192;
  const Real PurpleR = 192;
  const Real PurpleG = 128;
  const Real PurpleB = 255;
  const Real RedR = 255;
  const Real RedG = 0;
  const Real RedB = 0;
  const Real WhiteR = 255;
  const Real WhiteG = 255;
  const Real WhiteB = 255;
  const Real YellowR = 255;
  const Real YellowG = 255;
  const Real YellowB = 0;
  const Real BlackR = 0;
  const Real BlackG = 0;
  const Real BlackB = 0;
  const Real OrangeR = 255;
  const Real OrangeG = 165;
  const Real OrangeB = 0;
  /* String indices */
  enum {EP = 0, CapShape, CapSurface, CapColor, Bruises, Odor, GillAttachment,
	GillSpacing, GillSize, GillColor, StalkShape, StalkRoot,
	StalkSurfaceAboveRing, StalkSurfaceBelowRing, StalkColorAboveRing,
	StalkColorBelowRing, VeilType, VeilColor, RingNumber, RingType,
	SporePrintColor, Population, Habitat};
  /* Component indices */
  enum {CapShapeBell = 0,
	CapShapeConical,
	CapShapeConvex,
	CapShapeFlat,
	CapShapeKnobbed,
	CapShapeSunken,
	CapSurfaceFibrous,
	CapSurfaceGrooves,
	CapSurfaceScaly,
	CapSurfaceSmooth,
	CapColorR,
	CapColorG,
	CapColorB,
	BruisesTF,
	OdorAlmond,
	OdorAnise,
	OdorCreosote,
	OdorFishy,
	OdorFoul,
	OdorMusty,
	/* OdorNone, */
	OdorPungent,
	OdorSpicy,
	GillAttachmentAttached,
	GillAttachmentDescending,
	GillAttachmentFree,
	GillAttachmentNotched,
	GillSpacingClose,
	GillSpacingCrowded,
	GillSpacingDistant,
	GillSizeBN,
	GillColorR,
	GillColorG,
	GillColorB,
	StalkShapeET,
	StalkRootBulbous,
	StalkRootClub,
	StalkRootCup,
	StalkRootEqual,
	StalkRootRhizomorphs,
	StalkRootRooted,
	/* StalkRootMissing */
	StalkSurfaceAboveRingFibrous,
	StalkSurfaceAboveRingScaly,
	StalkSurfaceAboveRingSilky,
	StalkSurfaceAboveRingSmooth,
	StalkSurfaceBelowRingFibrous,
	StalkSurfaceBelowRingScaly,
	StalkSurfaceBelowRingSilky,
	StalkSurfaceBelowRingSmooth,
	StalkColorAboveRingR,
	StalkColorAboveRingG,
	StalkColorAboveRingB,
	StalkColorBelowRingR,
	StalkColorBelowRingG,
	StalkColorBelowRingB,
	VeilTypePU,
	VeilColorR,
	VeilColorG,
	VeilColorB,
	RingNumberOTT,
	RingTypeCobwebby,
	RingTypeEvanescent,
	RingTypeFlaring,
	RingTypeLarge,
	RingTypeNone,
	RingTypePendant,
	RingTypeSheathing,
	RingTypeZone,
	SporePrintColorR,
	SporePrintColorG,
	SporePrintColorB,
	PopulationAbundant,
	PopulationClustered,
	PopulationNumerous,
	PopulationScattered,
	PopulationSeveral,
	PopulationSolitary,
	HabitatGrasses,
	HabitatLeaves,
	HabitatMeadows,
	HabitatPaths,
	HabitatUrban,
	HabitatWaste,
	HabitatWoods
  };
  const size_t K = 2;		/* edible & poisonous */
  const size_t M = 8124;
  const size_t N1 = 22; /* Nominal components */
  const size_t N =
    6 + 		/* b, c, x, f, k, s = cap-shape*/
    4 + 		/* f, g, y, s = cap-surface */
    3 +		        /* n, b, c, g, r, p, u, e, w, y = cap-color */
    1 +			/* t, f = bruises? */
    8 +			/* a, l, c, y, f, m, n, p, s = odor */
    4 +			/* a, d, f, n = gill-attachment */
    3 +			/* c, w, d = gill-spacing */
    1 +			/* b, n = gill-size */
    3 +		        /* k, n, b, h, g, r, o, p, u, e, w, y = gill-color */
    1 +			/* e, t = stalk-shape */
    6 +			/* b, c, u, e, z, r, ? = stalk-root */
    4 +			/* f, y, k, s = stalk-surface-above-ring */
    4 +			/* f, y, k, s = stalk-surface-below-ring */
    3 +			/* n, b, c, g, o, p, e, w, y = stalk-color-above-ring */
    3 +			/* n, b, c, g, o, p, e, w, y = stalk-color-below-ring */
    1 +			/* p, u = veil-type */
    3 +			/* n, o, w, y = veil-color */
    1 +			/* n, o, t = ring-number */
    8 +			/* c, e, f, l, n, p, s, z = ring-type */
    3 +			/* k, n, b, h, r, o, u, w, y = spore-print-color */
    6 +			/* a, c, n, s, v, y = population */
    7;			/* g, l, m, p, u, w, d = habitat */
  const size_t K00 = 4208;	/* no. of edibles */
  const size_t K01 = 3916;	/* no. of poisonous */
  assert (Habitat == N1);
  assert (K00+K01 == M);
  assert (HabitatWoods == N-1);

  Real * const X = malloc (N * M * sizeof (Real));

  { /* Read in X */
    FILE *file = fopen ("agaricus-lepiota.data", "r");

    size_t i;
    for (i = 0; i < N * M; i++) X[i] = 0;

    char cs[2*N1+2];		/* for commas and terminating null */

    size_t cole = 0;		/* start column for edible */
    size_t colp = K00;		/* start column for poisonous */
    size_t ned = 0;		/* number of edibles read in */
    size_t npo = 0;		/* number of poisonous read in */

    for (i = 0; i < M; i++) {

      fscanf (file, "%s\n", cs);

      size_t col;

      switch (cs[EP]) {
      case 'e': col = cole; cole++; ned++; break;
      case 'p': col = colp; colp++; npo++; break;
      default: assert (false);
      };

      assert (col < M);
      assert (cole <= K00);
      assert (colp <= K00+K01);

      switch (cs[2*CapShape]) {
      case 'b': X[col * N + CapShapeBell] = 1; break;
      case 'c': X[col * N + CapShapeConical] = 1; break;
      case 'x': X[col * N + CapShapeConvex] = 1; break;
      case 'f': X[col * N + CapShapeFlat] = 1; break;
      case 'k': X[col * N + CapShapeKnobbed] = 1; break;
      case 's': X[col * N + CapShapeSunken] = 1; break;
      default: assert (false);
      };

      switch (cs[2*CapSurface]) {
      case 'f': X[col * N + CapSurfaceFibrous] = 1; break;
      case 'g': X[col * N + CapSurfaceGrooves] = 1; break;
      case 'y': X[col * N + CapSurfaceScaly] = 1; break;
      case 's': X[col * N + CapSurfaceSmooth] = 1; break;
      default: assert (false);
      };

      switch (cs[2*CapColor]) {
      case 'n':
	X[col * N + CapColorR] = BrownR / 255;
	X[col * N + CapColorG] = BrownG / 255;
	X[col * N + CapColorB] = BrownB / 255;
	break;
      case 'b':
	X[col * N + CapColorR] = BuffR / 255;
	X[col * N + CapColorG] = BuffG / 255;
	X[col * N + CapColorB] = BuffB / 255;
	break;
      case 'c':
	X[col * N + CapColorR] = CinnamonR / 255;
	X[col * N + CapColorG] = CinnamonG / 255;
	X[col * N + CapColorB] = CinnamonB / 255;
	break;
      case 'g':
	X[col * N + CapColorR] = GrayR / 255;
	X[col * N + CapColorG] = GrayG / 255;
	X[col * N + CapColorB] = GrayB / 255;
	break;
      case 'r':
	X[col * N + CapColorR] = GreenR / 255;
	X[col * N + CapColorG] = GreenG / 255;
	X[col * N + CapColorB] = GreenB / 255;
	break;
      case 'p':
	X[col * N + CapColorR] = PinkR / 255;
	X[col * N + CapColorG] = PinkG / 255;
	X[col * N + CapColorB] = PinkB / 255;
	break;
      case 'u':
	X[col * N + CapColorR] = PurpleR / 255;
	X[col * N + CapColorG] = PurpleG / 255;
	X[col * N + CapColorB] = PurpleB / 255;
	break;
      case 'e':
	X[col * N + CapColorR] = RedR / 255;
	X[col * N + CapColorG] = RedG / 255;
	X[col * N + CapColorB] = RedB / 255;
	break;
      case 'w':
	X[col * N + CapColorR] = WhiteR / 255;
	X[col * N + CapColorG] = WhiteG / 255;
	X[col * N + CapColorB] = WhiteB / 255;
	break;
      case 'y':
	X[col * N + CapColorR] = YellowR / 255;
	X[col * N + CapColorG] = YellowG / 255;
	X[col * N + CapColorB] = YellowB / 255;
	break;
      default: assert (false);
      }; /* end of switch (cs[2*CapColor]) */

      switch (cs[2*Bruises]) {
      case 't': X[col * N + BruisesTF] = 1; break;
      case 'f': X[col * N + BruisesTF] = 0; break;
      default: assert (false);
      };

      switch (cs[2*Odor]) {
      case 'a': X[col * N + OdorAlmond] = 1; break;
      case 'l': X[col * N + OdorAnise] = 1; break;
      case 'c': X[col * N + OdorCreosote] = 1; break;
      case 'y': X[col * N + OdorFishy] = 1; break;
      case 'f': X[col * N + OdorFoul] = 1; break;
      case 'm': X[col * N + OdorMusty] = 1; break;
      case 'n': /* X[col * N + OdorNone] = 1; */ break;
      case 'p': X[col * N + OdorPungent] = 1; break;
      case 's': X[col * N + OdorSpicy] = 1; break;
      default: assert (false);
      }; /* end of switch (cs[2*Odor]) */

      switch (cs[2*GillAttachment]) {
      case 'a': X[col * N + GillAttachmentAttached] = 1; break;
      case 'd': X[col * N + GillAttachmentDescending] = 1; break;
      case 'f': X[col * N + GillAttachmentFree] = 1; break;
      case 'n': X[col * N + GillAttachmentNotched] = 1; break;
      default: assert (false);
      }; /* End of switch (cs[2*GillAttachment]) */

      switch (cs[2*GillSpacing]) {
      case 'c': X[col * N + GillSpacingClose] = 1; break;
      case 'w': X[col * N + GillSpacingCrowded] = 1; break;
      case 'd': X[col * N + GillSpacingDistant] = 1; break;
      default: assert (false);
      };

      switch (cs[2*GillSize]) {
      case 'b': X[col * N + GillSizeBN] = 1; break;
      case 'n': X[col * N + GillSizeBN] = 0; break;
      default: assert (false);
      };

      switch (cs[2*GillColor]) {
      case 'k':
	X[col * N + GillColorR] = BlackR / 255;
	X[col * N + GillColorG] = BlackG / 255;
	X[col * N + GillColorB] = BlackB / 255;
	break;
      case 'n':
	X[col * N + GillColorR] = BrownR / 255;
	X[col * N + GillColorG] = BrownG / 255;
	X[col * N + GillColorB] = BrownB / 255;
	break;
      case 'b':
	X[col * N + GillColorR] = BuffR / 255;
	X[col * N + GillColorG] = BuffG / 255;
	X[col * N + GillColorB] = BuffB / 255;
	break;
      case 'h':
	X[col * N + GillColorR] = ChocolateR / 255;
	X[col * N + GillColorG] = ChocolateG / 255;
	X[col * N + GillColorB] = ChocolateB / 255;
	break;
      case 'g':
	X[col * N + GillColorR] = GrayR / 255;
	X[col * N + GillColorG] = GrayG / 255;
	X[col * N + GillColorB] = GrayB / 255;
	break;
      case 'r':
	X[col * N + GillColorR] = GreenR / 255;
	X[col * N + GillColorG] = GreenG / 255;
	X[col * N + GillColorB] = GreenB / 255;
	break;
      case 'o':
	X[col * N + GillColorR] = OrangeR / 255;
	X[col * N + GillColorG] = OrangeG / 255;
	X[col * N + GillColorB] = OrangeB / 255;
	break;
      case 'p':
	X[col * N + GillColorR] = PinkR / 255;
	X[col * N + GillColorG] = PinkG / 255;
	X[col * N + GillColorB] = PinkB / 255;
	break;
      case 'u':
	X[col * N + GillColorR] = PurpleR / 255;
	X[col * N + GillColorG] = PurpleG / 255;
	X[col * N + GillColorB] = PurpleB / 255;
	break;
      case 'e':
	X[col * N + GillColorR] = RedR / 255;
	X[col * N + GillColorG] = RedG / 255;
	X[col * N + GillColorB] = RedB / 255;
	break;
      case 'w':
	X[col * N + GillColorR] = WhiteR / 255;
	X[col * N + GillColorG] = WhiteG / 255;
	X[col * N + GillColorB] = WhiteB / 255;
	break;
      case 'y':
	X[col * N + GillColorR] = YellowR / 255;
	X[col * N + GillColorG] = YellowG / 255;
	X[col * N + GillColorB] = YellowB / 255;
	break;
      default: assert (false);
      };  /* End of switch (cs[2*GillColor]) */

      switch (cs[2*StalkShape]) {
      case 'e': X[col * N + StalkShapeET] = 1; break;
      case 't': X[col * N + StalkShapeET] = 0; break;
      default: assert (false);
      };

      switch (cs[2*StalkRoot]) {
      case 'b': X[col * N + StalkRootBulbous] = 1; break;
      case 'c': X[col * N + StalkRootClub] = 1; break;
      case 'u': X[col * N + StalkRootCup] = 1; break;
      case 'e': X[col * N + StalkRootEqual] = 1; break;
      case 'z': X[col * N + StalkRootRhizomorphs] = 1; break;
      case 'r': X[col * N + StalkRootRooted] = 1; break;
      case '?': break;
      default: assert (false);
      };

      switch (cs[2*StalkSurfaceAboveRing]) {
      case 'f': X[col * N + StalkSurfaceAboveRingFibrous] = 1; break;
      case 'y': X[col * N + StalkSurfaceAboveRingScaly] = 1; break;
      case 'k': X[col * N + StalkSurfaceAboveRingSilky] = 1; break;
      case 's': X[col * N + StalkSurfaceAboveRingSmooth] = 1; break;
      default: assert (false);
      };

      switch (cs[2*StalkSurfaceBelowRing]) {
      case 'f': X[col * N + StalkSurfaceBelowRingFibrous] = 1; break;
      case 'y': X[col * N + StalkSurfaceBelowRingScaly] = 1; break;
      case 'k': X[col * N + StalkSurfaceBelowRingSilky] = 1; break;
      case 's': X[col * N + StalkSurfaceBelowRingSmooth] = 1; break;
      default: assert (false);
      };

      switch (cs[2*StalkColorAboveRing]) {
      case 'n':
	X[col * N + StalkColorAboveRingR] = BrownR / 255;
	X[col * N + StalkColorAboveRingG] = BrownG / 255;
	X[col * N + StalkColorAboveRingB] = BrownB / 255;
	break;
      case 'b':
	X[col * N + StalkColorAboveRingR] = BuffR / 255;
	X[col * N + StalkColorAboveRingG] = BuffG / 255;
	X[col * N + StalkColorAboveRingB] = BuffB / 255;
	break;
      case 'c':
	X[col * N + StalkColorAboveRingR] = CinnamonR / 255;
	X[col * N + StalkColorAboveRingG] = CinnamonG / 255;
	X[col * N + StalkColorAboveRingB] = CinnamonB / 255;
	break;
      case 'g':
	X[col * N + StalkColorAboveRingR] = GrayR / 255;
	X[col * N + StalkColorAboveRingG] = GrayG / 255;
	X[col * N + StalkColorAboveRingB] = GrayB / 255;
	break;
      case 'o':
	X[col * N + StalkColorAboveRingR] = OrangeR / 255;
	X[col * N + StalkColorAboveRingG] = OrangeG / 255;
	X[col * N + StalkColorAboveRingB] = OrangeB / 255;
	break;
      case 'p':
	X[col * N + StalkColorAboveRingR] = PinkR / 255;
	X[col * N + StalkColorAboveRingG] = PinkG / 255;
	X[col * N + StalkColorAboveRingB] = PinkB / 255;
	break;
      case 'e':
	X[col * N + StalkColorAboveRingR] = RedR / 255;
	X[col * N + StalkColorAboveRingG] = RedG / 255;
	X[col * N + StalkColorAboveRingB] = RedB / 255;
	break;
      case 'w':
	X[col * N + StalkColorAboveRingR] = WhiteR / 255;
	X[col * N + StalkColorAboveRingG] = WhiteG / 255;
	X[col * N + StalkColorAboveRingB] = WhiteB / 255;
	break;
      case 'y':
	X[col * N + StalkColorAboveRingR] = YellowR / 255;
	X[col * N + StalkColorAboveRingG] = YellowG / 255;
	X[col * N + StalkColorAboveRingB] = YellowB / 255;
	break;
      default: assert (false);
      };

      switch (cs[2*StalkColorBelowRing]) {
      case 'n':
	X[col * N + StalkColorBelowRingR] = BrownR / 255;
	X[col * N + StalkColorBelowRingG] = BrownG / 255;
	X[col * N + StalkColorBelowRingB] = BrownB / 255;
	break;
      case 'b':
	X[col * N + StalkColorBelowRingR] = BuffR / 255;
	X[col * N + StalkColorBelowRingG] = BuffG / 255;
	X[col * N + StalkColorBelowRingB] = BuffB / 255;
	break;
      case 'c':
	X[col * N + StalkColorBelowRingR] = CinnamonR / 255;
	X[col * N + StalkColorBelowRingG] = CinnamonG / 255;
	X[col * N + StalkColorBelowRingB] = CinnamonB / 255;
	break;
      case 'g':
	X[col * N + StalkColorBelowRingR] = GrayR / 255;
	X[col * N + StalkColorBelowRingG] = GrayG / 255;
	X[col * N + StalkColorBelowRingB] = GrayB / 255;
	break;
      case 'o':
	X[col * N + StalkColorBelowRingR] = OrangeR / 255;
	X[col * N + StalkColorBelowRingG] = OrangeG / 255;
	X[col * N + StalkColorBelowRingB] = OrangeB / 255;
	break;
      case 'p':
	X[col * N + StalkColorBelowRingR] = PinkR / 255;
	X[col * N + StalkColorBelowRingG] = PinkG / 255;
	X[col * N + StalkColorBelowRingB] = PinkB / 255;
	break;
      case 'e':
	X[col * N + StalkColorBelowRingR] = RedR / 255;
	X[col * N + StalkColorBelowRingG] = RedG / 255;
	X[col * N + StalkColorBelowRingB] = RedB / 255;
	break;
      case 'w':
	X[col * N + StalkColorBelowRingR] = WhiteR / 255;
	X[col * N + StalkColorBelowRingG] = WhiteG / 255;
	X[col * N + StalkColorBelowRingB] = WhiteB / 255;
	break;
      case 'y':
	X[col * N + StalkColorBelowRingR] = YellowR / 255;
	X[col * N + StalkColorBelowRingG] = YellowG / 255;
	X[col * N + StalkColorBelowRingB] = YellowB / 255;
	break;
      default: assert (false);
      };

      switch (cs[2*VeilType]) {
      case 'p': X[col * N + VeilTypePU] = 1; break;
      case 'u': X[col * N + VeilTypePU] = 0; break;
      default: assert (false);
      };

      switch (cs[2*VeilColor]) {
      case 'n':
	X[col * N + VeilColorR] = BrownR / 255;
	X[col * N + VeilColorG] = BrownR / 255;
	X[col * N + VeilColorB] = BrownR / 255;
	break;
      case 'o':
	X[col * N + VeilColorR] = OrangeR / 255;
	X[col * N + VeilColorG] = OrangeR / 255;
	X[col * N + VeilColorB] = OrangeR / 255;
	break;
      case 'w':
	X[col * N + VeilColorR] = WhiteR / 255;
	X[col * N + VeilColorG] = WhiteR / 255;
	X[col * N + VeilColorB] = WhiteR / 255;
	break;
      case 'y':
	X[col * N + VeilColorR] = YellowR / 255;
	X[col * N + VeilColorG] = YellowR / 255;
	X[col * N + VeilColorB] = YellowR / 255;
	break;
      default: assert (false);
      };

      switch (cs[2*RingNumber]) {
      case 'n': X[col * N + RingNumberOTT] = 0.0; break;
      case 'o': X[col * N + RingNumberOTT] = 0.5; break;
      case 't': X[col * N + RingNumberOTT] = 1.0; break;
      default: assert (false);
      };

      switch (cs[2*RingType]) {
      case 'c': X[col * N + RingTypeCobwebby] = 1; break;
      case 'e': X[col * N + RingTypeEvanescent] = 1; break;
      case 'f': X[col * N + RingTypeFlaring] = 1; break;
      case 'l': X[col * N + RingTypeLarge] = 1; break;
      case 'n': X[col * N + RingTypeNone] = 1; break;
      case 'p': X[col * N + RingTypePendant] = 1; break;
      case 's': X[col * N + RingTypeSheathing] = 1; break;
      case 'z': X[col * N + RingTypeZone] = 1; break;
      default: assert (false);
      };

      switch (cs[2*SporePrintColor]) {
      case 'k':
	X[col * N + SporePrintColorR] = BlackR / 255;
	X[col * N + SporePrintColorG] = BlackG / 255;
	X[col * N + SporePrintColorB] = BlackB / 255;
	break;
      case 'n':
	X[col * N + SporePrintColorR] = BrownR / 255;
	X[col * N + SporePrintColorG] = BrownG / 255;
	X[col * N + SporePrintColorB] = BrownB / 255;
	break;
      case 'b':
	X[col * N + SporePrintColorR] = BuffR / 255;
	X[col * N + SporePrintColorG] = BuffG / 255;
	X[col * N + SporePrintColorB] = BuffB / 255;
	break;
      case 'h':
	X[col * N + SporePrintColorR] = ChocolateR / 255;
	X[col * N + SporePrintColorG] = ChocolateG / 255;
	X[col * N + SporePrintColorB] = ChocolateB / 255;
	break;
      case 'r':
	X[col * N + SporePrintColorR] = GreenR / 255;
	X[col * N + SporePrintColorG] = GreenG / 255;
	X[col * N + SporePrintColorB] = GreenB / 255;
	break;
      case 'o':
	X[col * N + SporePrintColorR] = OrangeR / 255;
	X[col * N + SporePrintColorG] = OrangeG / 255;
	X[col * N + SporePrintColorB] = OrangeB / 255;
	break;
      case 'u':
	X[col * N + SporePrintColorR] = PurpleR / 255;
	X[col * N + SporePrintColorG] = PurpleG / 255;
	X[col * N + SporePrintColorB] = PurpleB / 255;
	break;
      case 'w':
	X[col * N + SporePrintColorR] = WhiteR / 255;
	X[col * N + SporePrintColorG] = WhiteG / 255;
	X[col * N + SporePrintColorB] = WhiteB / 255;
	break;
      case 'y':
	X[col * N + SporePrintColorR] = YellowR / 255;
	X[col * N + SporePrintColorG] = YellowG / 255;
	X[col * N + SporePrintColorB] = YellowB / 255;
	break;
      default: assert (false);
      }; /* End of switch (cs[2*SporePrintColor]) */

      switch (cs[2*Population]) {
      case 'a': X[col * N + PopulationAbundant] = 1; break;
      case 'c': X[col * N + PopulationClustered] = 1; break;
      case 'n': X[col * N + PopulationNumerous] = 1; break;
      case 's': X[col * N + PopulationScattered] = 1; break;
      case 'v': X[col * N + PopulationSeveral] = 1; break;
      case 'y': X[col * N + PopulationSolitary] = 1; break;
      default: assert (false);
      };

      switch (cs[2*Habitat]) {
      case 'g': X[col * N + HabitatGrasses] = 1; break;
      case 'l': X[col * N + HabitatLeaves] = 1; break;
      case 'm': X[col * N + HabitatMeadows] = 1; break;
      case 'p': X[col * N + HabitatPaths] = 1; break;
      case 'u': X[col * N + HabitatUrban] = 1; break;
      case 'w': X[col * N + HabitatWaste] = 1; break;
      case 'd': X[col * N + HabitatWoods] = 1; break;
      default: assert (false);
      };

    };

    int err = fclose (file);
    assert (err == 0);

    assert (ned == K00);
    assert (npo == K01);
  };

  const size_t Ks = 2;
  const size_t L1 = K * Ks;
  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  /* fIMP parameters */
  const Real chi1 = (Real) M / ((Real) L1); /* M/L */
  const Real chi = 0.00001 * chi1;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  const Real gamma = 0.015;	/* O(1) */
  const Real alpha = 2.0001 * (L1 - 1) * beta;	/* M/L */
  const size_t Q = K * (2 + M / 10);
  const size_t MaxIter = 300;
  size_t maxIter = MaxIter; /* Return value of iteration count is stored here */

  size_t K1;			/* Returned value of K */
  K1 = fIMP (N, M, L1, X, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, &maxIter);
  printf ("Iterations = %d\n", (int) maxIter);
  printf ("True K = %d computed K = %d\n", (int) K, (int) K1);

  size_t L = 0;
  { /* Compute and print L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
    printf ("True L = %d computed L = %d\n", (int) L1, (int) L);
  };

  const size_t LambdaX[2] = {K00, K01};
  size_t wrongY;
  { /* Score based on Y */
    size_t wrongYC, wrongYR;
    wrongY = computeClassificationErrorsG (K1, K, Lambda, N, LambdaX, X, Y,
					   &wrongYC, &wrongYR);
    printf ("Wrong for computed Y = %d out of %d\n", (int) wrongY, (int) M);
    printf ("Wrong C = %d Wrong R = %d\n", (int) wrongYC, (int) wrongYR);
  };

  { /* Standard K-Means */
    maxIter = MaxIter;		/* reset */
    kmeans (N, X, M, Y, K, Q, &maxIter);
    printf ("kmeans took %d iterations\n", (int) maxIter);
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    const size_t Lambda1[2] = {1, 1};
    size_t wrongKC, wrongKR;
    wrongK = computeClassificationErrorsG (K, K, Lambda1, N, LambdaX, X, Y,
					   &wrongKC, &wrongKR);
    printf ("Wrong for K-means = %d out of %d\n", (int) wrongK, (int) M);
    printf ("Wrong C = %d Wrong R = %d\n", (int) wrongKC, (int) wrongKR);
  };

  free (X);
  free (Y);
  free (Lambda);
  free (Omega);
    
}; /* End of test5 */


/* *** *** */


/* Union of globular clusters */
void test6 (void) {

  const size_t K = 4;		/* No. of main clusters */
  const size_t Ks = 3;		/* No. of subclusters per cluster */
  const size_t N = 2;		/* Dimension of space */
  const size_t M1 = 5;		/* Samples per subcluster */
  const size_t M = K * Ks * M1;	/* Total no. of samples */

  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);

  Real * const shiftsMajor = malloc (N * K * sizeof (Real));
  Real * const shiftsMinor = malloc (N * (K * Ks) * sizeof (Real));
  assert (shiftsMajor != NULL);
  assert (shiftsMinor != NULL);

  /* Return random number in [-b,b] */
  inline Real randomReal (Real b) {
    return b * (2 * drand48 () - 1);
  };

  /* Inflation parameters for clusters */
  const Real nuMajor = 10.0 * LOG2(K) * POW (K, 1.0/N);
  const Real nuMinor = (nuMajor > 1) ? SQRT(nuMajor) : SQ(nuMajor);
  printf ("nuMajor = %g nuMinor = %g\n", (double) nuMajor, (double) nuMinor);

  { /* Set up shiftsMajor */
    randArray (shiftsMajor, N * K); /* random shifts on [-1,1]^N */
    size_t k;
    for (k = 0; k < K; k++)
      scaleColumn (N, shiftsMajor, k, randomReal (nuMajor));
  };

  { /* Set up shifts minor */
    randArray (shiftsMinor, N * K * Ks);
    size_t k;
    for (k = 0; k < K * Ks; k++)
      scaleColumn (N, shiftsMinor, k, randomReal (nuMinor));
    for (k = 0; k < K; k++)
      shiftMatrix (shiftsMajor + k * N, shiftsMinor + k * N * Ks, N, Ks);
  };

  { /* Set up X */
    randArray (X, N * M);
    size_t k;
    for (k = 0; k < K * Ks; k++)
      shiftMatrix (shiftsMinor + k * N, X + k * N * M1, N, M1);
  };

  /* To be fair scramble X */
  Real * const Xs = malloc (N * M * sizeof (Real));
  assert (Xs != NULL);
  /* Xs = X */
  memcpy (Xs, X, N * M * sizeof (Real));
  /* Scramble columns of Xs */
  randPerm (N * sizeof (Real), Xs, M);


  /* Set up for fIMP */
  const size_t Ky = 2;		/* No. of subclusters per cluster for Y */
  const size_t L1 = K * Ky;	/* No. of columns of Y */
  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  const Real chi1 = (Real) M / ((Real) L1); /* M/L */
  const Real chi = 0.00001 * chi1;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  const Real gamma = 0.001;	/* O(1) */
  const Real alpha = 2.0001 * (L1 - 1) * beta;	/* M/L */
  const size_t Q = K * (2 + M / 10);
  const size_t MaxIter = 300;
  size_t maxIter = MaxIter; /* Return value of iteration count is stored here */

  size_t K1;			/* Returned value of K */
  K1 = fIMP (N, M, L1, Xs, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, &maxIter);
  printf ("Iterations = %d\n", (int) maxIter);
  printf ("True K = %d computed K = %d\n", (int) K, (int) K1);

  size_t L = 0;
  { /* Compute and print L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
    printf ("True L = %d computed L = %d\n", (int) L1, (int) L);
  };

  size_t * const LambdaX = malloc (K * sizeof (size_t));
  { /* Set up LambdaX */
    size_t k;
    for (k = 0; k < K; k++) LambdaX[k] = Ks * M1;
  };

  size_t wrongY;
  { /* Score based on Y */
    size_t wrongYC, wrongYR;
    wrongY = computeClassificationErrorsG (K1, K, Lambda, N, LambdaX, X, Y,
					   &wrongYC, &wrongYR);
    printf ("Wrong for computed Y = %d out of %d\n", (int) wrongY, (int) M);
    printf ("Wrong C = %d Wrong R = %d\n", (int) wrongYC, (int) wrongYR);
  };

  { /* Standard K-Means */
    maxIter = MaxIter;		/* reset */
    kmeans (N, Xs, M, Y, K, Q, &maxIter);
    printf ("kmeans took %d iterations\n", (int) maxIter);
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    size_t k;
    for (k = 0; k < K; k++) Lambda1[k] = 1;
    size_t wrongKC, wrongKR;
    wrongK = computeClassificationErrorsG (K, K, Lambda1, N, LambdaX, X, Y,
					   &wrongKC, &wrongKR);
    printf ("Wrong for K-means = %d out of %d\n", (int) wrongK, (int) M);
    printf ("Wrong C = %d Wrong R = %d\n", (int) wrongKC, (int) wrongKR);
    free (Lambda1);
  };

  /* Visualize clusters */
  if (N == 2) {
    FILE *gplotout;

    gplotout = popen ("gnuplot -persist", "w");
    if (gplotout != NULL) {	/* We have gnuplot */

      /* Gnuplot color table */
      char *gpc[21];
      gpc[0] = "black";
      gpc[1] = "red";
      gpc[2] = "brown";
      gpc[3] = "green";
      gpc[4] = "blue";
      gpc[5] = "gray";
      gpc[6] = "greenyellow";
      gpc[7] = "yellow";
      gpc[8] = "plum";
      gpc[9] = "bisque";
      gpc[10] = "violet";
      gpc[11] = "orange";
      gpc[12] = "pink";
      gpc[13] = "honeydew";
      gpc[14] = "cyan";
      gpc[15] = "magenta";
      gpc[16] = "gold";
      gpc[17] = "salmon";
      gpc[18] = "khaki";
      gpc[19] = "beige";
      gpc[20] = "olive";
      
      /* *** *** */

      /* Set up gnuplot */
      bool pdf = false;
      if (pdf) {
	fprintf (gplotout, "set terminal pdf color\n");
	fprintf (gplotout, "set output \"fIMP.pdf\"\n");
      };
      fprintf (gplotout, "set multiplot\n");
      fprintf (gplotout, "set parametric\n");
      fprintf (gplotout, "set xrange [-%g:%g]\n", nuMajor+1, nuMajor+1);
      fprintf (gplotout, "set yrange [-%g:%g]\n", nuMajor+1, nuMajor+1);
      //fprintf (gplotout, "set size ratio -1\n");
      fprintf (gplotout, "unset xtics\n");
      fprintf (gplotout, "unset ytics\n");
      fprintf (gplotout, "set title \"fIMP\"\n");

      size_t k, j;

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "set object circle at %g,%g radius 0.05\n",
		 MREF (shiftsMajor, N, 0, k), MREF (shiftsMajor, N, 1, k));
      };

      { /* Draw computed centers */
	size_t ind = 0;
	for (k = 0; k < K1; k++) {
	  size_t l;
	  /* printf ("%d \"%s\" sub-cluster(s)\n", (int) Lambda[k], gpc[k]); */
	  for (l = 0; l < Lambda[k]; l++) {
	    fprintf (gplotout,
		     "set object rectangle at %g,%g size 0.1, 0.1 fc rgb \"%s\"\n",
		     MREF (Y, N, 0, ind), MREF (Y, N, 1, ind), gpc[k]);
	    ind++;
	  };
	};
      };

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "plot '-' w p lt %d\n", (int) (k+1));
	for (j = 0; j < Ks * M1; j++) {
	  fprintf (gplotout, "%g %g\n",
		   (double) MREF (X, N, 0, j+k*Ks*M1),
		   (double) MREF (X, N, 1, j+k*Ks*M1));
	};
	fprintf (gplotout, "e\n");
      };

      if (pdf) fprintf (gplotout, "set output\n");
      fflush (gplotout);
      
    };
  }; /* End of visualize clusters */

  /* Free memory */
  free (X);
  free (shiftsMajor);
  free (shiftsMinor);
  free (Xs);
  free (Y);
  free (Lambda);
  free (Omega);
  free (LambdaX);

}; /* End of test6 */


/* *** *** */

/* Primarily visualization in 2D.
   Same as test1 with one cluster worth of outliers.
   N = 2, gamma = 2E-3 for small K, Ks = 3
   N = 3, gamma = 8E-4, K = 100, Ks = 2
 */
void test7 (void) {

  size_t N = 3;
  size_t M1 = 30;			/* samples per cluster */
  size_t K = 100;			/* actual number of clusters */
  size_t KK = K+M1;		/* K for K-means */
  size_t M = M1 * (K + 1);	/* no. of samples; one "cluster" of outliers */
  size_t Ks = 2;			/* sub-clusters per cluster */
  size_t L1 = K * Ks;;
  assert (KK <= L1);

  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);
  randArray (X, N * M);		/* uniform on [-1,1)^N */

  /* Return random number in [-b,b] */
  inline Real randomReal (Real b) {
    return b * (2 * drand48 () - 1);
  };

  // const Real nu = 1.5 * K;	  /* Inflation parameter for clusters */
  const Real nu = 10.0 * LOG2(K) * POW(K, 1.0 / N);
  printf ("nu = %g\n", (double) nu);
  Real * const shifts = malloc (N * K * sizeof (Real));  /* True shifts */
  assert (shifts != NULL);

  { /* Set up shifts and shift clusters */

    randArray (shifts, N * K);	/* Random shifts uniform on [-1,1)^N */

    /* Random scaling & shift */
    size_t k;
    for (k = 0; k < K; k++) {
      scaleColumn (N, shifts, k, randomReal (nu)); /* random scaling */
      shiftMatrix (shifts + k * N, X + k * N * M1, N, M1);
    };

    //printMatrix (shifts, N, K);
  };

  { /* Scale outliers (last M1 columns of X) to [-nu,nu]^N */
    size_t k;
    for (k = M1 * K; k < M; k++)
      scaleColumn (N, X, k, nu);
  };

  Real * const Xs = malloc (N * M * sizeof (Real));
  /* Xs = X */
  memcpy (Xs, X, N * M * sizeof (Real));
  /* Scramble columns of Xs */
  randPerm (N * sizeof (Real), Xs, M);


  Real * const Y = malloc (N * L1 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  /* fIMP parameters */
  /* const Real gamma = 1E-3;
  const Real chi = 0.03;
  const Real beta = chi / (2.00001 * (L1 - 1));
  const Real alpha = 0.3 / (LOG2(N) * SQRT(K * Ks)); */
  const Real chi1 = (Real) M1 / ((Real) Ks); /* M/L */
  const Real chi = 0.001 * chi1;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  //const Real gamma = chi1 * (Real) Ks / (M1 * 2.0);
  const Real gamma = 8E-4;	/* O(1) */
  //const Real alpha = 0.0001 * (Real) M1 / ((Real) Ks);
  //const Real alpha = 0.001 * chi1;
  const Real alpha = 2 * (L1 - 1) * beta;	/* M/L */
  const size_t Q = 3; /* K * (2 + M1 / 5); */
  const size_t MaxIter = 300;
  size_t maxIter = MaxIter; /* Return value of iteration count is stored here */

  size_t K1;			/* Returned value of K */
  K1 = fIMP (N, M, L1, Xs, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, &maxIter);
  printf ("Iterations = %d\n", (int) maxIter);
  printf ("True K = %d computed K = %d\n", (int) K, (int) K1);

  size_t L = 0;
  { /* Compute and print L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
    printf ("True L = %d computed L = %d\n", (int) L1, (int) L);
  };

  size_t wrongY;
  { /* Score based on Y */
    /* Note that we skip the outliers for computing errors */
    wrongY = computeClassificationErrors (K1, K, Lambda, N, M1, X, Y);
    printf ("Wrong for computed Y = %d out of %d\n", (int) wrongY, (int) M);
  };

  size_t wrongS;
  { /* Score based on shifts */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    /* Ignore outliers for computing errors */
    wrongS = computeClassificationErrors (K, K, Lambda1, N, M1, X, shifts);
    printf ("Wrong for true shifts = %d out of %d\n", (int) wrongS, (int) M);
    free (Lambda1);
  };

  printf ("Final score = %g\n", ((double) wrongS - wrongY) / (2 * M));

  /* Visualize clusters */
  if (N == 2) {
    FILE *gplotout;

    gplotout = popen ("gnuplot -persist", "w");
    if (gplotout != NULL) {	/* We have gnuplot */

      /* Gnuplot color table */
      char *gpc[21];
      gpc[0] = "black";
      gpc[1] = "red";
      gpc[2] = "brown";
      gpc[3] = "green";
      gpc[4] = "blue";
      gpc[5] = "gray";
      gpc[6] = "greenyellow";
      gpc[7] = "yellow";
      gpc[8] = "plum";
      gpc[9] = "bisque";
      gpc[10] = "violet";
      gpc[11] = "orange";
      gpc[12] = "pink";
      gpc[13] = "honeydew";
      gpc[14] = "cyan";
      gpc[15] = "magenta";
      gpc[16] = "gold";
      gpc[17] = "salmon";
      gpc[18] = "khaki";
      gpc[19] = "beige";
      gpc[20] = "olive";
      
      /* *** *** */

      /* Set up gnuplot */
      bool pdf = false;
      if (pdf) {
	fprintf (gplotout, "set terminal pdf color\n");
	fprintf (gplotout, "set output \"fIMP.pdf\"\n");
      };
      fprintf (gplotout, "set multiplot\n");
      fprintf (gplotout, "set parametric\n");
      fprintf (gplotout, "set xrange [-%g:%g]\n", nu+1, nu+1);
      fprintf (gplotout, "set yrange [-%g:%g]\n", nu+1, nu+1);
      //fprintf (gplotout, "set size ratio -1\n");
      fprintf (gplotout, "unset xtics\n");
      fprintf (gplotout, "unset ytics\n");
      fprintf (gplotout, "set title \"fIMP\"\n");

      size_t k, j;

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "set object circle at %g,%g radius 0.05\n",
		 MREF (shifts, N, 0, k), MREF (shifts, N, 1, k));
      };

      { /* Draw computed centers */
	size_t ind = 0;
	for (k = 0; k < K1; k++) {
	  size_t l;
	  /* printf ("%d \"%s\" sub-cluster(s)\n", (int) Lambda[k], gpc[k]); */
	  for (l = 0; l < Lambda[k]; l++) {
	    fprintf (gplotout,
		     "set object rectangle at %g,%g size 0.1, 0.1 fc rgb \"%s\"\n",
		     MREF (Y, N, 0, ind), MREF (Y, N, 1, ind), gpc[k]);
	    ind++;
	  };
	};
      };

      /* Plot outliers too */
      for (k = 0; k < K+1; k++) {
	fprintf (gplotout, "plot '-' w p lt %d\n", (int) (k+1));
	for (j = 0; j < M1; j++) {
	  fprintf (gplotout, "%g %g\n",
		   (double) MREF (X, N, 0, j+k*M1),
		   (double) MREF (X, N, 1, j+k*M1));
	};
	fprintf (gplotout, "e\n");
      };

      if (pdf) fprintf (gplotout, "set output\n");
      fflush (gplotout);
      
    };
  }; /* End of visualize clusters */

  { /* Standard K-Means with Q = KK */
    maxIter = MaxIter;		/* reset */
    assert (KK <= L1);
    kmeans (N, Xs, M, Y, KK, KK, &maxIter);
    printf ("kmeans took %d iterations\n", (int) maxIter);
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (KK * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < KK; k++) Lambda1[k] = 1;
    };
    /* Ignore outliers */
    wrongK = computeClassificationErrors (KK, K, Lambda1, N, M1, X, Y);
    printf ("Wrong for K-means = %d out of %d\n", (int) wrongK, (int) M);
    free (Lambda1);
  };

  printf ("Final score for K-means = %g\n",
	  ((double) wrongS - wrongK) / (2 * M));

  /* Visualize kmeans clusters */
  if (N == 2) {
    FILE *gplotout;

    gplotout = popen ("gnuplot -persist", "w");
    if (gplotout != NULL) {	/* We have gnuplot */

      /* Gnuplot color table */
      char *gpc[21];
      gpc[0] = "black";
      gpc[1] = "red";
      gpc[2] = "brown";
      gpc[3] = "green";
      gpc[4] = "blue";
      gpc[5] = "gray";
      gpc[6] = "greenyellow";
      gpc[7] = "yellow";
      gpc[8] = "plum";
      gpc[9] = "bisque";
      gpc[10] = "violet";
      gpc[11] = "orange";
      gpc[12] = "pink";
      gpc[13] = "honeydew";
      gpc[14] = "cyan";
      gpc[15] = "magenta";
      gpc[16] = "gold";
      gpc[17] = "salmon";
      gpc[18] = "khaki";
      gpc[19] = "beige";
      gpc[20] = "olive";
      
      /* *** *** */

      /* Set up gnuplot */
      bool pdf = false;
      if (pdf) {
	fprintf (gplotout, "set terminal pdf color\n");
	fprintf (gplotout, "set output \"kmeans.pdf\"\n");
      };
      fprintf (gplotout, "set multiplot\n");
      fprintf (gplotout, "set parametric\n");
      fprintf (gplotout, "set xrange [-%g:%g]\n", nu+1, nu+1);
      fprintf (gplotout, "set yrange [-%g:%g]\n", nu+1, nu+1);
      //fprintf (gplotout, "set size ratio -1\n");
      fprintf (gplotout, "unset xtics\n");
      fprintf (gplotout, "unset ytics\n");
      fprintf (gplotout, "set title \"kmeans\"\n");

      size_t k, j;

      for (k = 0; k < K; k++) {
	fprintf (gplotout, "set object circle at %g,%g radius 0.05\n",
		 MREF (shifts, N, 0, k), MREF (shifts, N, 1, k));
      };

      /* Draw computed centers */
      for (k = 0; k < KK; k++) {
	fprintf (gplotout,
		 "set object rectangle at %g,%g size 0.1, 0.1 fc rgb \"%s\"\n",
		 MREF (Y, N, 0, k), MREF (Y, N, 1, k), gpc[k]);
      };

      /* Plot outliers too */
      for (k = 0; k < K+1; k++) {
	fprintf (gplotout, "plot '-' w p lt %d\n", (int) (k+1));
	for (j = 0; j < M1; j++) {
	  fprintf (gplotout, "%g %g\n",
		   (double) MREF (X, N, 0, j+k*M1),
		   (double) MREF (X, N, 1, j+k*M1));
	};
	fprintf (gplotout, "e\n");
      };

      if (pdf) fprintf (gplotout, "set output\n");
      fflush (gplotout);
      
    };
  }; /* End of visualize kmeans clusters */

  /*  Memory clean up */
  free (X);
  free (Xs);
  free (shifts);
  free (Y);
  free (Lambda);
  free (Omega);

}; /* End of test7 */


/* *** *** */

/* driver modified to allow outliers. */
/* Automatically choose Q=K for K-means runs */
void driver2 (const size_t N, const size_t K, const size_t M1, const size_t Ks,
	      const Real gamma, const Real chi, const Real alpha,
	      const size_t Q, const size_t MaxIter,
	      Real * const restrict scoreIMP,
	      Real * const restrict scoreKMeans,
	      size_t * const restrict impCount,
	      size_t * const restrict kmeansCount,
	      Real * timeIMP,
	      Real * timeKMeans) {
  
  size_t M = M1 * (K+1);	/* no. of samples including outliers */
  size_t L1 = K * Ks;		/* max number of sub-clusters */
  size_t KK = K + M1;		/* No. of clusters for K-means */
  // KK = KK <= L1 ? KK : L1;
  size_t L2 = KK <= L1 ? L1 : KK;
  assert (KK <= L2);		/* Make sure we have enough space */

  Real * const X = malloc (N * M * sizeof (Real));
  assert (X != NULL);
  randArray (X, N * M);

  /* Return random number in [-b,b] */
  inline Real randomReal (Real b) {
    return b * (2 * drand48 () - 1);
  };

  //const Real nu = 1.5 * K;  /* Inflation parameter for clusters */
  const Real nu = 10.0 * LOG2(K) * POW (K, 1.0/N);
  Real * const shifts = malloc (N * K * sizeof (Real));  /* True shifts */
  assert (shifts != NULL);

  { /* Set up shifts and shift clusters */
    randArray (shifts, N * K);	/* Random shifts */

    /* Random scaling & shift */
    size_t k;
    for (k = 0; k < K; k++) {
      scaleColumn (N, shifts, k, randomReal (nu)); /* random scaling */
      shiftMatrix (shifts + k * N, X + k * N * M1, N, M1);
    };

    //printMatrix (shifts, N, K);
  };

  { /* Scale outliers (last M1 columns of X) to [-nu,nu]^N */
    size_t k;
    for (k = M1 * K; k < M; k++)
      scaleColumn (N, X, k, nu);
  };

  Real * const Xs = malloc (N * M * sizeof (Real));
  /* Xs = X */
  memcpy (Xs, X, N * M * sizeof (Real));
  /* Scramble columns of Xs */
  randPerm (N * sizeof (Real), Xs, M);


  Real * const Y = malloc (N * L2 * sizeof (Real));
  size_t * const Lambda = malloc (L1 * sizeof (size_t));
  Real * const Omega = malloc (N * sizeof (Real));

  assert (Y != NULL);
  assert (Lambda != NULL);
  assert (Omega != NULL);

  { /* Set up Omega as avg(X) */
    size_t i;
    for (i = 0; i < N; i++) {
      size_t j;
      Omega[i] = 0;
      for (j = 0; j < M; j++) {
	Omega[i] += X[j * N + i];
      };
      Omega[i] /= M;
    };
  };

  Real beta = chi / (2.00001 * (L1 - 1));
  *impCount = MaxIter;	/* Return value is actual count */
  size_t K1;			/* Returned value of K */
  clock_t start, end;
  start = clock ();
  K1 = fIMP (N, M, L1, Xs, Lambda, Y,
	     alpha, beta, gamma, chi, Omega, Q, impCount);
  end = clock ();
  *timeIMP = ((double) (end - start)) / CLOCKS_PER_SEC;
  size_t L = 0;
  { /* Compute L */
    size_t k;
    for (k = 0; k < K1; k++) L += Lambda[k];
  };

  /* Score based on Y */
  size_t wrongY;
  wrongY = computeClassificationErrors (K1, K, Lambda, N, M1, X, Y);

  size_t wrongS;
  { /* Score based on shifts */
    size_t * const Lambda1 = malloc (K * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < K; k++) Lambda1[k] = 1;
    };
    wrongS = computeClassificationErrors (K, K, Lambda1, N, M1, X, shifts);
    free (Lambda1);
  };

  *scoreIMP = ((double) wrongS - wrongY) / (2 * M);

  { /* Standard K-Means using Q = KK*/
    clock_t start, end;
    *kmeansCount = MaxIter;
    start = clock ();
    kmeans (N, Xs, M, Y, KK, KK, kmeansCount);
    end = clock ();
    *timeKMeans = ((double) (end - start)) / CLOCKS_PER_SEC;
  };

  size_t wrongK;
  { /* Score based on kmeans centers */
    size_t * const Lambda1 = malloc (KK * sizeof (size_t));
    {
      size_t k;
      for (k = 0; k < KK; k++) Lambda1[k] = 1;
    };
    wrongK = computeClassificationErrors (KK, K, Lambda1, N, M1, X, Y);
    free (Lambda1);
  };

  *scoreKMeans = ((double) wrongS - wrongK) / (2 * M);

  /* Free memory */
  free (X);
  free (shifts);
  free (Xs);
  free (Y);
  free (Lambda);
  free (Omega);

}; /* End of driver2 */


/* *** *** */

/* test2 with driver2 for M1 outlier experiments */
/* Gather error statistics. Repeatedly run tests and gather statistics
   on scores and iterations counts. */
void test8 () {

  /* nu = 10.0 * LOG2(K) * POW(K,1/N)

     K = 100, Ks = 2 Q = 30 KK = K + M1
     N = 7, gamma = 4E-4 chi = 0.001 * chi1 M1 = 30
     N = 14, gamma = 4E-4 chi = 0.001 * chi1 M1 = 30
     N = 28, gamma = 4E-4, chi = 0.001 * chi1 M1 = 60
     N = 56, gamma = 3E-4, chi = 0.001 * chi1 M1 = 120
     N = 112, gamma = 3E-4, chi = 0.001 * chi1 M1 = 240

     K = 200, Ks = 2 Q = 40
     N = 7, gamma = 2E-4 chi = 0.001 * chi1 M1 = 30
     N = 14, gamma = 2E-4 chi = 0.001 * chi1 M1 = 30
     N = 28, gamma = 2E-4, chi = 0.001 * chi1 M1 = 60
     N = 56, gamma = 2E-4, chi = 0.001 * chi1 M1 = 120
     N = 112, gamma = 1E-4, chi = 0.001 * chi1 M1 = 240
 */

  /* Fixed parameters of test */
  const size_t Trials = 10;		/* Number of trials */
  const size_t N = 7;			/* dimension */
  const size_t K = 100;			/* number of clusters */
  const size_t M1 = 30;			/* Samples per cluster */
  const size_t Ks = 2;			/* Sub-clusters per cluster */
  const Real chi1 = (Real) M1 / Ks;	/* M/L */
  const Real chi = 0.001 * chi1;
  const Real gamma = 4E-4;	/* O(1) */
  // const Real gamma = 3E-4;
  const size_t L1 = K * Ks;
  const Real beta = chi / (2.00001 * (L1 - 1)); /* M/L^2 */
  const Real alpha = 2 * (L1 - 1) * beta;		/* M/L */
  // const Real alpha = 5E-3;
  const size_t Q = 30; /* K * (2 + M1 / 5); */
  const size_t MaxIter = 300;

  /* Output variables for driver2 */
  Real scoreIMP = 0;
  Real scoreKMeans = 0;
  size_t impCount = MaxIter;
  size_t kmeansCount = MaxIter;
  Real timeIMP;
  Real timeKMeans;

  size_t trial;
  size_t wins = 0, absWins = 0;
  size_t losses = 0, absLosses = 0;
  size_t ties = 0, absTies = 0;
  Real avgScoreIMP = 0;
  Real avgScoreKMeans = 0;
  Real avgTimeIMP = 0;
  Real avgTimeKMeans = 0;
  for (trial = 0; trial < Trials; trial++) {
    printf ("\b\b\b\b\b\b\b\b\b\b\b\b\b\btrial = %d", (int) trial);
    fflush (stdout);
    driver2 (N, K, M1, Ks, gamma, chi, alpha, Q, MaxIter,
	     &scoreIMP, &scoreKMeans, &impCount, &kmeansCount,
	     &timeIMP, &timeKMeans);
    if (impCount >= MaxIter) printf ("IMP requires more iterations\n");
    if (kmeansCount >= MaxIter) printf ("KMeans requires more iterations\n");
    if (scoreIMP > scoreKMeans)
      wins++;
    else if (scoreIMP < scoreKMeans)
      losses++;
    else
      ties++;
    if (scoreIMP > 0)
      absWins++;
    else if (scoreIMP < 0)
      absLosses++;
    else
      absTies++;
    avgScoreIMP += scoreIMP;
    avgScoreKMeans += scoreKMeans;
    avgTimeIMP += timeIMP;
    avgTimeKMeans += timeKMeans;
  };

  printf ("\nN = %d Q = %d M1 = %d K = %d Ks = %d Trials = %d\n",
	  (int) N, (int) Q, (int) M1, (int) K, (int) Ks, (int) Trials);
  printf ("chi = %g alpha = %g beta = %g gamma = %g\n",
	  (double) chi, (double) alpha, (double) beta, (double) gamma);
  printf ("AbsWins = %d AbsTies = %d AbsLosses = %d\n",
	  (int) absWins, (int) absTies, (int) absLosses);
  printf ("Wins = %d Ties = %d Losses = %d\n",
	  (int) wins, (int) ties, (int) losses);
  printf ("Avg. score of IMP = %g\n", (double) avgScoreIMP / Trials);
  printf ("Avg. score of KMeans = %g\n", (double) avgScoreKMeans / Trials);
  printf ("Avg. time of IMP = %g\n", (double) avgTimeIMP / Trials);
  printf ("Avg. time of KMeans = %g\n", (double) avgTimeKMeans / Trials);

}; /* End of test8 */

/* *** *** */

int main (void) {
  srand (time (0));
  srand48 (time (0));

  // test1 (); printf ("\n");
  // test2 (); printf ("\n");
  // test3 (); printf ("\n");
  // test4 (); printf ("\n");
  // test5 (); printf ("\n");
  // test6 (); printf ("\n");
  // test7 (); printf ("\n");
  test8 (); printf ("\n");

  return EXIT_SUCCESS;
}; /* End of main */

/* *** *** */
